'''Tests for auton_lr.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,duplicate-code,protected-access,too-many-locals
# pylint: disable=too-many-lines

from copy import deepcopy
import pickle
import socket
import time
from typing import Tuple

import numpy as np
import pandas as pd
import pytest
from sklearn import datasets  # type: ignore[import]


from ...config_components.distributed_config import DistributedConfig
from ...conftest import Clearer, Waiter
from ...metrics.impl.metric_auto import MetricCatalogAuto
from ...problem_def.task import TaskType
from ...tables.impl.table_auto import TableCatalogAuto
from ...wrangler.base_port import BasePort
from ...wrangler.dataset import Column, Dataset, Metadata, RoleName
from ...wrangler.logger import Logger

from ..impl.algorithm_auto import AlgorithmCatalogAuto
from ..impl.synchronous import advance, read_from_neighbors
from ..sklearn.sklearn_algorithm import SklearnAlgorithm

from .auton_logistic_regression import (
    AutonLogisticRegression, AutonLogisticRegressionInstance, AutonLogisticRegressionNeighbor)

_ = TableCatalogAuto()
log = Logger(__file__, to_stdout=False).logger()

base_port = BasePort()


CORRECT_V = np.array([-0.5086, 2.3477, -0.2656, 19.2694])

CORRECT_V_L2 = 10.0
CORRECT_DIST_PARAMS = {
    'L2': CORRECT_V_L2,
    'Lambda': 100,
    'omega': 0.6
}
CORRECT_V_ACC = 0.86
GOOD_ENOUGH_ACC = 0.85


def load_classification_dataset() -> Tuple[Dataset, Dataset, Dataset, Dataset]:
    # Load the breast cancer dataset
    bc_x_full, bc_y_series = datasets.load_breast_cancer(return_X_y=True, as_frame=True)
    assert isinstance(bc_x_full, pd.DataFrame)
    assert isinstance(bc_y_series, pd.Series)
    bc_y = pd.DataFrame({'target': bc_y_series})

    # restrict number of attributes for wider variability of results
    bc_x = bc_x_full.iloc[:, :3]

    test_size = 50
    # Split the data into training/testing sets
    bc_x_train = bc_x[:-test_size]
    bc_x_test = bc_x[-test_size:]

    # Split the targets into training/testing sets
    bc_y_train = bc_y[:-test_size]
    bc_y_test = bc_y[-test_size:].reset_index()

    metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        task=TaskType.BINARY_CLASSIFICATION
    )

    dataset_train = Dataset(
        metadata=metadata,
        covariates=bc_x_train,
        target=bc_y_train
    )

    dataset_test = Dataset(
        metadata=metadata,
        covariates=bc_x_test
    )
    ground_truth = Dataset(
        metadata=metadata,
        ground_truth=bc_y_test
    )

    train_cov_0 = dataset_train['covariates'][np.array(dataset_train['target']) == 0]
    train_cov_1 = dataset_train['covariates'][np.array(dataset_train['target']) == 1]
    reduced_train = Dataset(
        covariates=pd.concat((train_cov_0.iloc[0:1], train_cov_1.iloc[0:1]), axis=0) * 0,
        target=pd.DataFrame({'target': [0] * 1 + [1] * 1}),
        metadata=metadata
    )

    return (dataset_train, dataset_test, ground_truth, reduced_train)


LOOPBACK = '127.0.0.1'
SUNNY_DAY_PORT = base_port.next()


def test_sunny_day(assert_no_exceptions: Clearer) -> None:
    '''Sunny day with no neighbors produces the same result as sklearn.'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '2': [],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('2', (LOOPBACK, SUNNY_DAY_PORT)),
                ]
            },
        },
        'my_id': 2,
    })

    sklearn_alg = AlgorithmCatalogAuto().lookup_by_name('sklearn.linear_model.LogisticRegression')
    alg = AutonLogisticRegression()

    # Setting Lambda and L2 values to make our results similar to
    #   sklearn's logistic regression, in order to be more certain
    #   that our algorithm is reasonable.
    dut = alg.instantiate(
        distributed=distributed_config,
        L2=CORRECT_V_L2)
    assert isinstance(dut, AutonLogisticRegressionInstance)

    sklearn_dut = sklearn_alg.instantiate(
        C=1 / CORRECT_V_L2,
        penalty='l2',
        fit_intercept=True,
    )

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth, _ = load_classification_dataset()

    try:
        dut.start()
        dut.fit(train)
        sklearn_dut.fit(train)

        # sklearn_dut_v = np.array(list(sklearn_dut.impl.coef_.ravel())
        #                          + [sklearn_dut.impl.intercept_[0]])
        # dut_v = dut.params

        result_train = dut.predict(train)
        sklearn_result_train = sklearn_dut.predict(train)
        sklearn_train_acc = metric.calculate(
            pred=sklearn_result_train,
            ground_truth=Dataset(metadata=train.metadata, ground_truth=train['target']))
        result_train_acc = metric.calculate(
            pred=result_train,
            ground_truth=Dataset(metadata=train.metadata, ground_truth=train['target']))

        assert result_train_acc == pytest.approx(expected=sklearn_train_acc, abs=0.01)

        result_test = dut.predict(test)
        sklearn_result_test = sklearn_dut.predict(test)
        sklearn_test_acc = metric.calculate(pred=sklearn_result_test,
                                            ground_truth=ground_truth)
        result_test_acc = metric.calculate(pred=result_test,
                                           ground_truth=ground_truth)

        assert result_test_acc == pytest.approx(expected=sklearn_test_acc, abs=0.01)
        assert result_test_acc == CORRECT_V_ACC
        assert_no_exceptions([dut])

    finally:
        dut.stop()


RECEIVE_EVENT_SENDER_PORT = base_port.next()
RECEIVE_EVENT_RECEIVER_PORT = base_port.next()


def test_receive_event(assert_no_exceptions: Clearer) -> None:
    '''Poor training is improved with information from a neighbor.'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                },
            },
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, RECEIVE_EVENT_SENDER_PORT)),
                    ('2', (LOOPBACK, RECEIVE_EVENT_RECEIVER_PORT)),
                ],
            },
        },
        'my_id': 2,
    })

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    _, test, ground_truth, reduced_train = load_classification_dataset()

    dut = AutonLogisticRegression().instantiate(distributed=distributed_config, synchronous=True)

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, RECEIVE_EVENT_SENDER_PORT))

    # These parameters were hand-extracted from a successful fit.
    message = pickle.dumps((
        1,
        CORRECT_V,
        None
    ))

    try:
        dut.start()
        dut.fit(reduced_train)
        advance([dut])
        check = dut.predict(test)

        # Confirm that the training is very bad.
        assert metric.calculate(pred=check,
                                ground_truth=ground_truth
                                ) == pytest.approx(expected=0.2, abs=0.01)

        sock.sendto(message, (LOOPBACK, RECEIVE_EVENT_RECEIVER_PORT))

        # Wait for the message to arrive and get processed.
        time.sleep(1.0)

        read_from_neighbors([dut])
        dut.fit(reduced_train)  # fit again on same data to incorporate info from neighbor

        advance([dut])
        # Confirm that we learned from our neighbor.
        got = dut.predict(test)

        assert metric.calculate(pred=got,
                                ground_truth=ground_truth
                                ) > GOOD_ENOUGH_ACC
        assert_no_exceptions([dut])
    finally:
        dut.stop()
        sock.close()


SEND_SENDER_PORT = base_port.next()
SEND_RECEIVER_PORT = base_port.next()


def test_send(assert_no_exceptions: Clearer) -> None:
    '''We share our coefficients with a neighbor.'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, SEND_SENDER_PORT)),
                    ('2', (LOOPBACK, SEND_RECEIVER_PORT)),
                ],
            },
        },
        'my_id': 1,
    })

    train, _, _, _ = load_classification_dataset()

    receiver = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    receiver.bind((LOOPBACK, SEND_RECEIVER_PORT))
    dut = None

    try:
        dut = AutonLogisticRegression().instantiate(distributed=distributed_config)
        dut.start()
        dut.fit(train)

        got_raw = receiver.recv(1024)  # Wait for the message to arrive and get processed.
        got = AutonLogisticRegressionNeighbor.decode(got_raw)

        assert got.v == pytest.approx(CORRECT_V, abs=0.001)
        assert_no_exceptions([dut])

    finally:
        receiver.close()
        if dut is not None:
            dut.stop()


TWO_SAME_RECEIVER_PORT = base_port.next()
TWO_SAME_SENDER_PORT = base_port.next()


def test_two_same(wait_til_all_fit: Waiter,
                  assert_no_exceptions: Clearer) -> None:
    '''Two nodes with the same training don't change their coefficients.'''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, TWO_SAME_RECEIVER_PORT)),
                    ('2', (LOOPBACK, TWO_SAME_SENDER_PORT)),
                ]
            }
        },
        'my_id': 0,
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth, _ = load_classification_dataset()

    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    alg = AutonLogisticRegression()
    dut_receiver = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause),
        **CORRECT_DIST_PARAMS)
    dut_sender = None
    try:
        dut_receiver.start()
        dut_receiver.fit(train)

        dut_sender = alg.instantiate(
            distributed=DistributedConfig(distributed_sender_clause),
            **CORRECT_DIST_PARAMS)
        dut_sender.start()
        dut_sender.fit(train)

        # Wait for the message to arrive and get processed.
        wait_til_all_fit([dut_receiver, dut_sender], convergence_check=True)

        # Confirm that neighbor optimization doesn't mess up result.
        got_receiver = dut_receiver.predict(test)
        got_sender = dut_sender.predict(test)

        assert metric.calculate(pred=got_receiver,
                                ground_truth=ground_truth
                                ) == CORRECT_V_ACC

        assert metric.calculate(pred=got_sender,
                                ground_truth=ground_truth
                                ) == CORRECT_V_ACC

        assert got_receiver is not None
        assert got_sender is not None

        pd.testing.assert_frame_equal(got_receiver.predictions, got_sender.predictions)

        assert dut_sender.params == pytest.approx(dut_receiver.params, abs=0.001)
        assert_no_exceptions([dut_sender, dut_receiver])
    finally:
        dut_receiver.stop()
        if dut_sender is not None:
            dut_sender.stop()


INTEGRATED_RECEIVER_PORT = base_port.next()
INTEGRATED_SENDER_PORT = base_port.next()


def test_integrated(assert_no_exceptions: Clearer) -> None:
    ''''Two nodes talking to each other.

    One gets poor taining, the other gets full training.
    We see that both nodes eventually get good results.
    '''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, INTEGRATED_RECEIVER_PORT)),
                    ('2', (LOOPBACK, INTEGRATED_SENDER_PORT)),
                ]
            }
        },
        'my_id': 0,
        'fit_eps': 0.01
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth, reduced_train = load_classification_dataset()
    # TODO(Dan) Catch the case where there is only 1 class in the training data.
    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    alg = AutonLogisticRegression()
    receiver = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause),
        synchronous=True,
        **CORRECT_DIST_PARAMS)

    dut = None
    try:
        receiver.start()
        receiver.fit(reduced_train)

        advance([receiver])

        check = receiver.predict(test)

        # Confirm that the training is bad.
        assert metric.calculate(pred=check,
                                ground_truth=ground_truth
                                ) < 0.8

        dut = alg.instantiate(
            distributed=DistributedConfig(distributed_sender_clause),
            synchronous=True,
            **CORRECT_DIST_PARAMS)
        dut.start()
        dut.fit(train)

        # Wait for the message to arrive and get processed.
        for _ in range(3):
            advance([dut, receiver])

        got = receiver.predict(test)

        # Actually 0.98! Is this a bug? It's better than the non-distributed result.
        assert metric.calculate(pred=got,
                                ground_truth=ground_truth
                                ) > GOOD_ENOUGH_ACC
        assert_no_exceptions([dut, receiver])
    finally:
        receiver.stop()
        if dut is not None:
            dut.stop()


TWO_HALF_RECEIVER_PORT = base_port.next()
TWO_HALF_SENDER_PORT = base_port.next()


def test_two_half(assert_no_exceptions: Clearer) -> None:
    '''Two nodes with half the training behaves as (reasonably) expected.'''
    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, TWO_HALF_RECEIVER_PORT)),
                    ('2', (LOOPBACK, TWO_HALF_SENDER_PORT)),
                ]
            }
        },
        'my_id': 0,
        'fit_eps': 0.001
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth, _ = load_classification_dataset()

    n = train['covariates'].shape[0]
    half_n = int(train['covariates'].shape[0] / 2)
    train_first = Dataset(
        metadata=train.metadata,
        covariates=train['covariates'].iloc[0:half_n],
        target=train['target'].iloc[0:half_n]
    )
    train_second = Dataset(
        metadata=train.metadata,
        covariates=train['covariates'].iloc[half_n:],
        target=train['target'].iloc[half_n:]
    )

    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    dist_params_first = dict(CORRECT_DIST_PARAMS)
    dist_params_first.update(
        L2=(half_n / n) * CORRECT_DIST_PARAMS['L2'],
        Lambda=10000000)
    dist_params_second = dict(dist_params_first)
    dist_params_second.update(L2=((n - half_n) / n) * CORRECT_DIST_PARAMS['L2'])

    alg = AutonLogisticRegression()
    dut_first = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause),
        synchronous=True,
        **dist_params_first)
    dut_second = None

    try:
        dut_first.start()
        dut_first.fit(train_first)

        advance([dut_first])

        dut_second = alg.instantiate(
            distributed=DistributedConfig(distributed_sender_clause),
            synchronous=True,
            **dist_params_second)
        dut_second.start()
        dut_second.fit(train_second)

        # Wait for the message to arrive and get processed.
        for _ in range(10):
            advance([dut_first, dut_second], min_time=1_000)

        # Confirm that neighbor optimization doesn't mess up result.
        got_receiver = dut_first.predict(test)
        got_sender = dut_second.predict(test)

        assert metric.calculate(pred=got_receiver,
                                ground_truth=ground_truth
                                ) > GOOD_ENOUGH_ACC

        assert metric.calculate(pred=got_sender,
                                ground_truth=ground_truth
                                ) > GOOD_ENOUGH_ACC

        assert got_receiver is not None
        assert got_sender is not None
        pd.testing.assert_frame_equal(got_receiver.predictions, got_sender.predictions)

        assert dut_second.params == pytest.approx(dut_first.params, rel=0.01)
        assert_no_exceptions([dut_first, dut_second])
    finally:
        dut_first.stop()
        if dut_second is not None:
            dut_second.stop()


NO_FIT_RECEIVER_PORT = base_port.next()
NO_FIT_SENDER_PORT = base_port.next()


def test_receive_no_fit(wait_til_all_fit: Waiter,
                        assert_no_exceptions: Clearer) -> None:
    '''Test that we spontaneously fit on a message from a neighbor without having seen data.'''
    distributed_config = DistributedConfig(clause={
        'polling_interval': '0.2',
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, NO_FIT_SENDER_PORT)),
                    ('2', (LOOPBACK, NO_FIT_RECEIVER_PORT)),
                ],
            },
        },
        'my_id': 2,
    })

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    _, test, ground_truth, _ = load_classification_dataset()

    dut = AutonLogisticRegression().instantiate(distributed=distributed_config)
    sorted_ground_truth = ground_truth.sorted_columns()

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, NO_FIT_SENDER_PORT))

    # These parameters were hand-extracted from a successful fit.
    message = pickle.dumps((
        1,
        CORRECT_V,
        None
    ))

    try:
        dut.start()
        sock.sendto(message, (LOOPBACK, NO_FIT_RECEIVER_PORT))

        # Wait for the message to arrive and get processed.
        wait_til_all_fit([dut])

        # Confirm that we learned from our neighbor.
        got = dut.predict(test)

        assert metric.calculate(pred=got,
                                ground_truth=sorted_ground_truth
                                ) > GOOD_ENOUGH_ACC
        assert_no_exceptions([dut])
    finally:
        dut.stop()
        sock.close()


TEST_N_SAME_N = 5
TEST_N_SAME_PORTS = [base_port.next() for _ in range(TEST_N_SAME_N)]


def test_n_same(wait_til_all_fit: Waiter,
                assert_no_exceptions: Clearer) -> None:
    '''N nodes fully connected with the same training match the non-distributed result.'''
    n = TEST_N_SAME_N

    adjacency = {f'{k}': [j for j in range(1, n + 1) if j != k] for k in range(1, n + 1)}
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_N_SAME_PORTS[k - 1])) for k in range(1, n + 1)]

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'my_id': 0,
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth, _ = load_classification_dataset()

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonLogisticRegression()

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc),
        **CORRECT_DIST_PARAMS) for dc in distributed_clauses]
    try:
        for dut in duts:
            dut.start()

        for dut in duts:
            dut.fit(train)

        # Wait for the message to arrive and get processed.
        wait_til_all_fit(duts, convergence_check=True)

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(test) for dut in duts]

        for got_prediction in got_predictions:
            assert metric.calculate(pred=got_prediction,
                                    ground_truth=ground_truth
                                    ) == pytest.approx(CORRECT_V_ACC, rel=0.05)

        gp0 = got_predictions[0]
        assert gp0 is not None
        for i in range(1, n):
            gpi = got_predictions[i]
            assert gpi is not None
            np.testing.assert_array_equal(gp0.predictions, gpi.predictions)

        for i in range(1, n):
            assert duts[0].params == pytest.approx(duts[i].params, rel=0.05)

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


TEST_TELEPHONE_N = 5
TEST_TELEPHONE_PORTS = [base_port.next() for _ in range(TEST_TELEPHONE_N)]


def test_telephone(wait_til_all_fit: Waiter,
                   assert_no_exceptions: Clearer) -> None:
    '''Whisper down the lane test. Nodes connected in a line.
    The only node that has data is one at the end '''
    n = TEST_TELEPHONE_N

    adjacency = {f'{k}': [k - 1, k + 1] for k in range(2, n)}
    adjacency['1'] = [2]
    adjacency[f'{n}'] = [n - 1]
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_TELEPHONE_PORTS[k - 1]))
                           for k in range(1, n + 1)]

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'my_id': 0,
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth, _ = load_classification_dataset()

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonLogisticRegression()

    telephone_params = dict(CORRECT_DIST_PARAMS, Lambda=1)
    telephone_params['omega'] = .5

    params = [dict(telephone_params) if k == 0 else dict(telephone_params) for k in range(n)]

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc),
        **ps) for (dc, ps) in zip(distributed_clauses, params)]
    try:
        for dut in duts:
            dut.start()

        duts[0].fit(train)

        # Wait for the message to arrive and get processed.
        assert wait_til_all_fit(duts, convergence_check=True, min_time=1.0, max_time=60.0), (
            'Timeout waiting for convergence')

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(test) for dut in duts]

        for got_prediction in got_predictions:
            assert metric.calculate(pred=got_prediction,
                                    ground_truth=ground_truth
                                    ) == CORRECT_V_ACC

        gp0 = got_predictions[0]
        assert gp0 is not None
        for i in range(1, n):
            gpi = got_predictions[i]
            assert gpi is not None
            np.testing.assert_array_equal(gp0.predictions, gpi.predictions)

        for i in range(n - 1):
            assert duts[i].params == pytest.approx(duts[i + 1].params, rel=0.01), (
                f'Node {i} and Node {i + 1} have different parameters'
            )

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


CONFLICT_RECEIVER_PORT = base_port.next()
CONFLICT_SENDER_PORT = base_port.next()


# TODO(Merritt/Piggy): investigate this.
@pytest.mark.skip('failing in CI but not locally.  Training gets worse when adding new data.')
def test_node_id_conflict_ok(wait_til_all_fit: Waiter,
                             assert_no_exceptions: Clearer) -> None:
    '''Nodes don't need to agree about which node has which ID.

    In this test, both nodes think they are Node 1.
    '''
    distributed_clause = {
        'polling_interval': '0.1',
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                ]
            }
        },
        'my_id': 1,  # This is the same for both nodes.
        'fit_eps': 0.01
    }

    receiver_clause = deepcopy(distributed_clause)
    assert isinstance(receiver_clause['communicator'], dict)  # to make mypy happy
    assert isinstance(receiver_clause['communicator']['sockets'], dict)  # to make mypy happy
    receiver_clause['communicator']['sockets']['nodes_and_endpoints'] = [
        ('1', (LOOPBACK, CONFLICT_RECEIVER_PORT))
    ]
    receiver_clause['discoverer'] = {
        'name': 'dynamic'
    }

    sender_clause = deepcopy(distributed_clause)
    assert isinstance(sender_clause['communicator'], dict)  # to make mypy happy
    assert isinstance(sender_clause['communicator']['sockets'], dict)  # to make mypy happy
    sender_clause['communicator']['sockets']['nodes_and_endpoints'] = [
        ('1', (LOOPBACK, CONFLICT_SENDER_PORT)),
        ('2', (LOOPBACK, CONFLICT_RECEIVER_PORT))
    ]
    sender_clause['discoverer'] = {
        'name': 'static',
        'static': {
            'adjacency': {
                '1': [2],
            }
        }
    }

    receiver_config = DistributedConfig(receiver_clause)
    sender_config = DistributedConfig(sender_clause)

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth, _ = load_classification_dataset()
    reduced_train = Dataset({
        'covariates': train['covariates'].iloc[516:517] * 0,
        'target': train['target'].iloc[516:517]
    }, metadata=train.metadata)

    alg = AutonLogisticRegression()
    receiver = alg.instantiate(distributed=receiver_config)
    sender = None
    try:
        receiver.start()
        receiver.fit(reduced_train)

        assert wait_til_all_fit([receiver], convergence_check=False), (
            'Timeout waiting for fit'
        )

        check = receiver.predict(test)

        # Confirm that the training is very bad.
        assert metric.calculate(pred=check,
                                ground_truth=ground_truth
                                ) == pytest.approx(expected=0.8, abs=0.01)

        sender = alg.instantiate(distributed=sender_config)
        sender.start()
        sender.fit(train)

        # Wait for the message to arrive and get processed.
        assert wait_til_all_fit([sender, receiver]), (
            'Timeout waiting for fit'
        )

        # Confirm that we learned from our neighbor.
        got = receiver.predict(test)

        assert metric.calculate(pred=got,
                                ground_truth=ground_truth
                                ) > 0.8
        assert_no_exceptions([sender, receiver])
    finally:
        receiver.stop()
        if sender is not None:
            sender.stop()


def load_multiclass_dataset() -> Tuple[Dataset, Dataset, Dataset]:
    # Load the iris dataset
    bc_x_full, bc_y_series = datasets.load_iris(return_X_y=True, as_frame=True)
    assert isinstance(bc_x_full, pd.DataFrame)
    assert isinstance(bc_y_series, pd.Series)
    bc_y = pd.DataFrame({'target': bc_y_series})

    # restrict number of attributes for wider variability of results
    bc_x = bc_x_full.iloc[:, :3]

    # shuffle our data
    bc_x = bc_x.sample(frac=1, random_state=1701)
    bc_y = bc_y.sample(frac=1, random_state=1701)

    test_size = 50
    # Split the data into training/testing sets
    bc_x_train = bc_x[:-test_size]
    bc_x_test = bc_x[-test_size:]

    # Split the targets into training/testing sets
    bc_y_train = bc_y[:-test_size]
    bc_y_test = bc_y[-test_size:]

    metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        task=TaskType.BINARY_CLASSIFICATION
    )

    dataset_train = Dataset(
        metadata=metadata,
        covariates=bc_x_train,
        target=bc_y_train
    )

    dataset_test = Dataset(
        metadata=metadata,
        covariates=bc_x_test
    )
    ground_truth = Dataset(
        metadata=metadata,
        ground_truth=bc_y_test
    )
    return (dataset_train, dataset_test, ground_truth)


@pytest.mark.skip(reason='Distributed logistic regression does not support multiclass yet.')
def test_multiclass(assert_no_exceptions: Clearer) -> None:
    '''Non-distributed multiclass classification.'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '2': [],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('2', (LOOPBACK, base_port.next())),
                ]
            },
        },
        'my_id': 2,
    })

    alg = AutonLogisticRegression()
    sklearn_alg = AlgorithmCatalogAuto().lookup_by_name('sklearn.linear_model.LogisticRegression')
    assert isinstance(sklearn_alg, SklearnAlgorithm)

    # Setting Lambda and L2 values to make our results similar to
    #   sklearn's logistic regression, in order to be more certain
    #   that our algorithm is reasonable.
    dut = alg.instantiate(
        distributed=distributed_config,
        L2=CORRECT_V_L2)
    sklearn_instance = sklearn_alg.instantiate(
        C=1 / CORRECT_V_L2,
        penalty='l2',
        fit_intercept=True,
    )

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth = load_multiclass_dataset()
    try:
        dut.start()
        dut.fit(train)
        sklearn_instance.fit(train)

        result = dut.predict(test)
        result_sklearn = sklearn_instance.predict(test)

        dut_acc = metric.calculate(pred=result,
                                   ground_truth=ground_truth)
        sklearn_acc = metric.calculate(pred=result_sklearn,
                                       ground_truth=ground_truth)
        assert dut_acc == pytest.approx(sklearn_acc, rel=.001)
        assert_no_exceptions([dut])
    finally:
        dut.stop()


TEST_N_SAME_N_DJAM = 5
TEST_N_SAME_PORTS_DJAM = [base_port.next() for _ in range(TEST_N_SAME_N_DJAM)]


def test_n_same_djam(assert_no_exceptions: Clearer) -> None:
    '''N nodes fully connected with the same training match the non-distributed result.'''
    n = TEST_N_SAME_N_DJAM

    adjacency = {f'{k}': [j for j in range(0, n) if j != k] for k in range(0, n)}
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_N_SAME_PORTS_DJAM[k])) for k in range(0, n)]

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'regularization': {
            'name': 'djam',
            'weight_matrix': [
                [0, 1, 1, 1, 1],
                [1, 0, 1, 1, 1],
                [1, 1, 0, 1, 1],
                [1, 1, 1, 0, 1],
                [1, 1, 1, 1, 0]
            ]
        },
        'my_id': 0,
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth, _ = load_classification_dataset()

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(0, n)]

    alg = AutonLogisticRegression()

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc),
        synchronous=True,
        **CORRECT_DIST_PARAMS) for dc in distributed_clauses]
    try:
        for dut in duts:
            dut.start()

        for dut in duts:
            dut.fit(train)

        # Wait for the message to arrive and get processed.
        advance(duts)

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(test) for dut in duts]

        for got_prediction in got_predictions:
            assert metric.calculate(pred=got_prediction,
                                    ground_truth=ground_truth
                                    ) == pytest.approx(CORRECT_V_ACC, rel=0.01)

        gp0 = got_predictions[0]
        assert gp0 is not None
        for i in range(1, n):
            gpi = got_predictions[i]
            assert gpi is not None
            np.testing.assert_array_equal(gp0.predictions, gpi.predictions)

        for i in range(1, n):
            assert duts[0].params == pytest.approx(duts[i].params, rel=0.05)

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


TEST_TELEPHONE_N_DJAM = 5
TEST_TELEPHONE_PORTS_DJAM = [base_port.next() for _ in range(TEST_TELEPHONE_N_DJAM)]


def test_telephone_djam(wait_til_all_fit: Waiter,
                        assert_no_exceptions: Clearer) -> None:
    '''Whisper down the lane test. Nodes connected in a line.
    The only node that has data is one at the end '''
    n = TEST_TELEPHONE_N_DJAM

    adjacency = {f'{k}': [k - 1, k + 1] for k in range(1, n - 1)}
    adjacency['0'] = [1]
    adjacency[f'{n-1}'] = [n - 2]
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_TELEPHONE_PORTS_DJAM[k]))
                           for k in range(0, n)]

    distributed_clause = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'regularization': {
            'name': 'djam',
            'weight_matrix': [
                [0, 1, 0, 0, 0],
                [1, 0, 1, 0, 0],
                [0, 1, 0, 1, 0],
                [0, 0, 1, 0, 1],
                [0, 0, 0, 1, 0]
            ]
        },
        'my_id': 0,
    }

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')

    train, test, ground_truth, _ = load_classification_dataset()

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(0, n)]

    alg = AutonLogisticRegression()

    telephone_params = dict(CORRECT_DIST_PARAMS, Lambda=1)
    telephone_params['omega'] = .5

    params = [dict(telephone_params) if k == 0 else dict(telephone_params) for k in range(n)]

    duts = [alg.instantiate(
        distributed=DistributedConfig(dc),
        **ps) for (dc, ps) in zip(distributed_clauses, params)]
    try:
        for dut in duts:
            dut.start()

        duts[0].fit(train)

        # Wait for the message to arrive and get processed.
        assert wait_til_all_fit(duts, convergence_check=True, min_time=1.0), (
            'Timeout waiting for convergence'
        )

        # Confirm that neighbor optimization doesn't mess up result.
        got_predictions = [dut.predict(test) for dut in duts]

        for got_prediction in got_predictions:
            assert metric.calculate(pred=got_prediction,
                                    ground_truth=ground_truth
                                    ) == CORRECT_V_ACC

        gp0 = got_predictions[0]
        assert gp0 is not None
        for i in range(1, n):
            gpi = got_predictions[i]
            assert gpi is not None
            np.testing.assert_array_equal(gp0.predictions, gpi.predictions)

        for i in range(n - 1):
            assert duts[i].params == pytest.approx(duts[i + 1].params, rel=0.01)

        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


COLS_SENDER_PORT = base_port.next()
COLS_RECEIVER_PORT = base_port.next()


def test_columns(assert_no_exceptions: Clearer) -> None:
    '''Test that columns are handled in data loading, sending and receiving.'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1],
                },
            },
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, COLS_SENDER_PORT)),
                    ('2', (LOOPBACK, COLS_RECEIVER_PORT)),
                ],
            },
        },
        'my_id': 2,
    })

    _, _, _, reduced_train = load_classification_dataset()

    dut = AutonLogisticRegression().instantiate(
        distributed=distributed_config, synchronous=True, **CORRECT_DIST_PARAMS)

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, COLS_SENDER_PORT))

    # These parameters were hand-extracted from a successful fit.
    match_message = pickle.dumps((
        1, CORRECT_V, reduced_train['covariates'].columns.to_list()
    ))

    mismatch_message = pickle.dumps((
        1, CORRECT_V, reduced_train['covariates'].columns.to_list() + ['hamster']
    ))

    try:
        dut.start()
        dut.fit(reduced_train)

        advance([dut])
        sock.sendto(match_message, (LOOPBACK, COLS_RECEIVER_PORT))

        advance([dut])
        assert_no_exceptions([dut])

        sock.sendto(mismatch_message, (LOOPBACK, COLS_RECEIVER_PORT))

        advance([dut])
        got = dut.poll_exceptions()
        assert got is not None
        assert isinstance(got[0], NotImplementedError)
        assert 'mismatch' in str(got[1])
        assert 'hamster' in str(got[1])

        # make sure there are no other exceptions we don't expect
        assert_no_exceptions([dut])

    finally:
        dut.stop()
        sock.close()


def test_roc(assert_no_exceptions: Clearer) -> None:
    '''Test bug where distributed logistic regression fails for ranking with roc'''
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '2': [],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('2', (LOOPBACK, base_port.next())),
                ]
            },
        },
        'my_id': 2,
    })

    alg = AutonLogisticRegression()

    dut = alg.instantiate(
        distributed=distributed_config)
    assert isinstance(dut, AutonLogisticRegressionInstance)
    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('roc_auc_score')

    train, test, ground_truth, _ = load_classification_dataset()

    try:
        dut.start()
        dut.fit(train)

        result = dut.predict(test)

        metric.calculate(
            pred=result,
            ground_truth=ground_truth)

        assert_no_exceptions([dut])

    finally:
        dut.stop()



'''Test the DataLoaderCatalog'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..stub_data_loader import DataLoaderStub
from .data_loader_catalog import DataLoaderCatalogStub

# pylint: disable=missing-function-docstring,duplicate-code


def test_register() -> None:
    dut = DataLoaderCatalogStub()
    assert 'stub' == dut.register(obj=DataLoaderStub)


def test_register_name() -> None:
    dut = DataLoaderCatalogStub()
    stub_data_loader = DataLoaderStub
    assert 'somenameddataloader' == dut.register(
        obj=stub_data_loader, name='SomeNamedDataLoader')


def test_lookup_by_name() -> None:
    dut = DataLoaderCatalogStub()
    some_data_loader = DataLoaderStub
    dut.register(obj=some_data_loader, name='SomeNamedDataLoader')
    assert dut.lookup_by_name('SomeNamedDataLoader') == some_data_loader


def test_lookup_one_by_tag() -> None:
    dut = DataLoaderCatalogStub()
    dut.register(obj=DataLoaderStub, name='NamedStubDataLoader', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 1


def test_lookup_by_tag_fail() -> None:
    dut = DataLoaderCatalogStub()
    dut.register(obj=DataLoaderStub, name='NamedStubDataLoader', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='different_value')) == 0


def test_lookup_multi_by_tag() -> None:
    dut = DataLoaderCatalogStub()
    stub_data_loader = DataLoaderStub
    dut.register(obj=stub_data_loader, name="Name1", tags={'k1': ['v1']})
    stub_data_loader2 = DataLoaderStub
    dut.register(obj=stub_data_loader2, name="Name2", tags={'k1': ['v1']})
    stub_data_loader3 = DataLoaderStub
    dut.register(obj=stub_data_loader3, name="Name3", tags={'k1': ['different_value']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 2

# TODO(Merritt): test for detection of tags in class variables as well as used when registering



'''Tests for metric_auto.py.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
from pathlib import Path

import pytest

from ...wrangler.constants import PACKAGE_NAME
from ..sklearn_metrics import SklearnMetric
from .metric import Metric
from .metric_auto import MetricCatalogAuto

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code


FAKE_METRIC_SOURCE = '''
from typing import Any, Optional
from ngautonml.metrics.impl.metric import Metric
from ngautonml.metrics.impl.metric_catalog import MetricCatalog

class FakeMetric(Metric):
    _name = 'fake_metric'
    _tags = {
        'some_tag': ['some_value']
    }

    def calculate(self, pred: Any, ground_truth: Optional[Any] = None) -> float:
        return 1.0


def register(catalog: MetricCatalog):
    catalog.register(FakeMetric())
'''


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = retval / f'.{PACKAGE_NAME}' / 'plugins' / 'metrics' / 'fake_metric.py'
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(FAKE_METRIC_SOURCE)
    return retval


def test_metric_catalog_auto() -> None:
    dut = MetricCatalogAuto()
    assert isinstance(dut.lookup_by_name('ROC_AUC_score'), SklearnMetric)
    assert len(dut.lookup_by_tag_and(task='binary_classification')) >= 2


def test_home_directory_auto(tmp_path) -> None:
    os.environ['HOME'] = str(tmp_path)
    dut = MetricCatalogAuto()
    got = dut.lookup_by_name('fake_metric')
    assert isinstance(got, Metric)
    assert got.tags['some_tag'] == ['some_value']



'''Test for issues #3 on gitlab (https://gitlab.com/autonlab/ngautonml/-/issues/3)

Essentially, just a test that ngautonml installs in a fresh environment.
'''
# pylint: disable=duplicate-code

from glob import glob
from pathlib import Path
import random
from shutil import rmtree
import subprocess
import sys

import pytest

from .virtual_module import VirtualModule


def test_bug3(tmp_path_factory: pytest.TempPathFactory) -> None:
    '''Build ngautonml.whl and make sure we can install it in a blank environment.'''

    # randomly skip this test 4 out of 5 times because it is very expensive.
    if random.randint(1, 5) != 5:
        return

    # Build ngautonml to create wheel file.
    ngautonml_root = Path(__file__).parents[2]
    rmtree(ngautonml_root / 'dist', ignore_errors=True)
    subprocess.run([sys.executable, '-m', 'poetry', 'build'], cwd=ngautonml_root, check=True)

    wheel_paths = glob(str(ngautonml_root / 'dist'
                           / 'ngautonml*py3-none-any.whl'))
    assert len(wheel_paths) == 1, (
        f'BUG: unexpected extra whl files: {wheel_paths}')
    wheel_path = Path(wheel_paths[0])

    # Create a new environment in a tempdir just for this test.
    tmpdir = tmp_path_factory.mktemp('bug3_test')
    tmp_module = VirtualModule(tmpdir)
    tmp_module.install_virtual_env()

    # Device Under Test
    # This will raise an error if not exit code 0
    tmp_module.install_under_venv(wheel_path)

    # Clean up the virtual environment.
    tmp_module.finalize()



'''Catalog for aggregators

   These can be used to select pipelines based on multiple unreliable metrics.
'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, List, Optional, Union

from ...catalog.memory_catalog import MemoryCatalog
from ...aggregators.impl.aggregator import Aggregator


class AggregatorCatalog(MemoryCatalog[Aggregator], metaclass=abc.ABCMeta):
    '''Base class for aggregator catalogs'''
    def register(self, obj: Aggregator, name: Optional[str] = None,
                 tags: Optional[Dict[str, Union[str, List[str]]]] = None) -> str:
        return super().register(obj, name, tags)


class AggregatorCatalogStub(AggregatorCatalog):
    '''stub'''



'''Tests for model_auto.py'''
# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code
from .algorithm_auto import AlgorithmCatalogAuto
from ..sklearn.sklearn_algorithm import SklearnAlgorithm
from ..connect import ConnectorModel
from ..extract_columns_by_role import ExtractColumnsByRoleModel
from ...problem_def.task import TaskType

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


def test_linear_regression() -> None:
    dut = AlgorithmCatalogAuto()
    model = dut.lookup_by_name('sklearn.linear_model.LinearRegression')
    assert isinstance(model, SklearnAlgorithm)
    items = dut.lookup_by_tag_and(task=TaskType.REGRESSION.name)
    assert any(isinstance(item, SklearnAlgorithm) for _, item in items.items())


def test_connector() -> None:
    dut = AlgorithmCatalogAuto()
    item = dut.lookup_by_name(name='connect')
    assert isinstance(item, ConnectorModel)


def test_extract() -> None:
    dut = AlgorithmCatalogAuto()
    model = dut.lookup_by_name(name='Extract Columns by Role')
    assert isinstance(model, ExtractColumnsByRoleModel)


def test_classifiers() -> None:
    dut = AlgorithmCatalogAuto()

    algorithms = dut.lookup_by_tag_and(task=TaskType.MULTICLASS_CLASSIFICATION.name)
    names = [algo.name for algo in algorithms.values()]
    assert set({'sklearn.linear_model.logisticregression',
                'sklearn.ensemble.randomforestclassifier',
                'sklearn.svm.svc',
                'sklearn.svm.linearsvc',
                'sklearn.naive_bayes.multinomialnb',
                'sklearn.discriminant_analysis.lineardiscriminantanalysis',
                'sklearn.discriminant_analysis.quadraticdiscriminantanalysis',
                'sklearn.ensemble.adaboostclassifier',
                'sklearn.tree.extratreeclassifier',
                'sklearn.ensemble.baggingclassifier',
                'sklearn.linear_model.passiveaggressiveclassifier',
                'sklearn.neural_network.mlpclassifier',
                'sklearn.ensemble.gradientboostingclassifier'}).issubset(set(names))



'''Base class for loaders that create pandas dataframes.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
import queue
import time
from typing import Any, Dict, Optional, List, Union

import pandas as pd
from sklearn.model_selection import KFold  # type: ignore[import]

from ...config_components.impl.config_component import (
    ConfigError, ValidationErrors, ParsingErrors)
from ...config_components.distributed_config import DistributedConfig
from ...problem_def.task import TaskType
from ...wrangler.constants import JSONKeys
from ...wrangler.dataset import Dataset, Column, RoleName, DatasetKeys
from ...config_components.dataset_config import DatasetFileError

from .data_loader import DataLoader


class Error(BaseException):
    '''Base class for all errors in this file.'''


class ColumnError(Error):
    '''Attempt to select columns in a dataframe that don't exist.'''


class DataframeLoader(DataLoader, metaclass=abc.ABCMeta):
    '''Base class for loaders that create pandas dataframes.'''
    _train_dataset: Dataset
    _que: queue.Queue

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._train_dataset = self._load_train()
        if self._train_dataset is not None:
            self._train_dataset.dataframe = self._str_column_names(self._train_dataset.dataframe)

        self._update_roles()

        self.validate(self._train_dataset, train=True)

        self._que = queue.Queue()
        self._que.put(self._train_dataset)

    def _update_roles(self) -> None:
        '''Give all unassigned columns the Attribute role.'''
        assigned_colnames: List[str] = []
        errors: List[ConfigError] = []

        if self._train_dataset is None:
            return

        for cols in self._metadata.roles.values():
            for col in cols:
                assigned_colnames.append(str(col.name))

        unassigned_colnames = list(
            set(self._train_dataset.dataframe.columns) - set(assigned_colnames))
        unassigned_cols = []
        for colname in unassigned_colnames:
            unassigned_cols.append(Column(colname))

        updated_roles = self._metadata.roles
        if RoleName.ATTRIBUTE not in updated_roles:
            updated_roles[RoleName.ATTRIBUTE] = []
        updated_roles[RoleName.ATTRIBUTE].extend(unassigned_cols)

        self._metadata = self._metadata.override_roles(updated_roles)

        # Because train data is already loaded, we need to update its
        # metadata separately.
        self._train_dataset._metadata = self._metadata  # pylint: disable=protected-access

        if len(errors) != 0:
            raise ParsingErrors(errors)

    def _prune_target(self, retval: Dataset) -> Dataset:
        '''remove target col if it is in data

        called by subclasses in load_test() since test data shouldn't have target

        keep target for forecasting problems
        '''
        # TODO(Merritt/Piggy): once we have a forecasting dataloader, we can
        #   do something more elegant than checking the task type here.

        if self._metadata.task != TaskType.FORECASTING:
            target = self._metadata.target
            if target is not None and target.name in retval.dataframe.columns:
                new_data = retval.dataframe.drop(
                    labels=target.name, axis=1)
                retval.dataframe = new_data

        return retval

    def _str_column_names(self, df: pd.DataFrame) -> pd.DataFrame:
        df.columns = df.columns.astype(str)
        return df

    def _ensure_cols_present(self, df: pd.DataFrame, role: RoleName) -> List[ConfigError]:
        errors: List[ConfigError] = []
        if role in self._metadata.roles:
            for col in self._metadata.roles[role]:
                if col.name not in df.columns:
                    errors.append(DatasetFileError(
                        f'Column {col.name} specified under role {role}'
                        ' not found among columns in dataset. '
                        f'Found: {df.columns}'))
        return errors

    def validate(self, dataset: Optional[Dataset], train: bool = False) -> None:
        '''Raise a ValidationErrors if something about the dataset is inconsistent.'''
        if dataset is None:
            return  # A None dataset is valid.
        errors: List[ConfigError] = []

        if DatasetKeys.DATAFRAME.value in dataset:
            errors.extend(self._ensure_cols_present(dataset.dataframe, RoleName.INDEX))
            if train:
                errors.extend(self._ensure_cols_present(dataset.dataframe, RoleName.ATTRIBUTE))
                errors.extend(self._ensure_cols_present(dataset.dataframe, RoleName.TARGET))
        elif DatasetKeys.GROUND_TRUTH.value in dataset:
            errors.extend(self._ensure_cols_present(dataset.ground_truth, RoleName.INDEX))
            errors.extend(self._ensure_cols_present(dataset.ground_truth, RoleName.TARGET))
        else:
            # Raise error?
            pass

        if len(errors) != 0:
            raise ValidationErrors(errors)

    def load_train(self) -> Optional[Dataset]:
        retval = self._train_dataset
        self.validate(retval, train=True)
        return self._conditional_split(retval)

    @abc.abstractmethod
    def _load_train(self) -> Optional[Dataset]:
        ...

    def load_test(self) -> Optional[Dataset]:
        '''Load test data if it exists, otherwise None'''
        retval = self._load_test()
        if retval is None:
            return None

        retval.dataframe = self._str_column_names(retval.dataframe)

        retval = self._prune_target(retval)
        self.validate(retval)
        # Unlike train data, don't split the test data for a distributed setting.
        return retval

    @abc.abstractmethod
    def _load_test(self) -> Optional[Dataset]:
        ...

    def _load_ground_truth(self) -> Optional[Dataset]:
        '''If the target column exists in the test set, extract it as the ground truth.'''
        test_set = self._load_test()
        if test_set is None:
            print('test set is None')
            return None

        if self._metadata.target is None:
            print('target in metadata is None')

            return None

        target_name = self._metadata.target.name

        if target_name in test_set.dataframe.columns:
            retval = Dataset(metadata=self._metadata)
            retval.ground_truth = test_set.dataframe[[target_name]]
            return retval

        print(f'target name ({target_name}) not in cols ({test_set.dataframe.columns}).')
        return None

    def _dataset(self,
                 data: Any,
                 key: Union[str, DatasetKeys] = DatasetKeys.DATAFRAME,
                 cols: Optional[List[str]] = None,
                 roles: Optional[List[Union[RoleName, str]]] = None,
                 **kwargs) -> Dataset:
        '''Load a Dataset object, by placing data at the supplied key.

        As of 2023-12-04 we only know how to handle things we can turn into a pandas DataFrame.

        Args:
          data:
            Dataframe or object that can be turned into one.
          key:
            The key for the data in the dataset. Defaults to "dataframe".
          cols:
            If set, only selects supplied column(s).
            Will take union with roles if both are specified.
          roles:
            If set, only selects columns with supplied role(s).
            Will take union with cols if both are specified.
        '''

        retval = Dataset(metadata=self._metadata)

        data_df = pd.DataFrame(data)
        data_df = self._str_column_names(data_df)

        # Trim to selected columns if either cols or roles is set.
        cols_to_select = set()
        if cols is not None:
            cols_to_select.update(cols)
        if roles is not None:
            for role in roles:
                if isinstance(role, str):
                    role = RoleName[role.upper()]
                cols_to_select.update(
                    c.name for c in self._metadata.roles[role])
        if cols_to_select:
            if cols_to_select.issubset(data_df.columns):
                data_df = data_df[sorted(cols_to_select)]
            else:
                raise ColumnError(
                    f'selecting columns {cols_to_select} from '
                    f'dataframe with columns {data_df.columns};'
                    'not all requested columns are present.'
                )

        if isinstance(key, str):
            key = DatasetKeys[key.upper()]

        retval[key.value] = data_df
        return retval

    def _conditional_split(self, data: Dataset) -> Dataset:
        '''Split a dataset if we are in a distributed simulation, otherwise do nothing.

        Requires 1 <= my_id <= num_nodes.
        '''
        problem_def = self._config.metadata.problem_def
        if problem_def is None:
            return data

        distributed = problem_def.get_conf('distributed')
        assert isinstance(distributed, DistributedConfig), (
            f'BUG: Expected distributed to be a DistributedConfig, '
            f'got {type(distributed)} instead.'
        )
        distributed_split = distributed.split
        if distributed_split is None:
            return data

        # split data for a distributed simulation
        kfcv = KFold(
            n_splits=distributed_split.num_nodes,
            shuffle=True,
            random_state=distributed_split.seed  # defaults to default seed
        )

        # 1-based: 1 <= my_id <= num_nodes.
        n = distributed.my_id
        # Choose nth split.
        _, indices = next(
            (x for i, x in enumerate(
                kfcv.split(data.dataframe),
                start=1
            ) if i == n), (None, None))
        assert indices is not None, (
            f'BUG: Could not find {n}th split of {distributed_split.num_nodes} '
            f'for node {distributed.my_id}'
        )

        split_df = data.dataframe.iloc[indices]
        split_df.reset_index(inplace=True, drop=True)
        retval = data.output()
        retval.dataframe = split_df
        return retval

    def poll(self, timeout: Optional[float] = 0.0) -> Optional[Dataset]:
        next_data = self._poll(timeout=timeout)
        if next_data is None:
            return None
        next_data.dataframe = self._str_column_names(next_data.dataframe)
        self.validate(next_data)

        return self._conditional_split(data=next_data)

    def _poll(self, timeout: Optional[float] = 0.0) -> Optional[Dataset]:
        '''Return the latest unsplit dataset.'''
        if timeout is not None:
            time.sleep(timeout)
        try:
            return self._que.get(block=False, timeout=timeout)
        except queue.Empty:
            return None

    def build_dataset_from_json(self, json_data: Dict) -> Dataset:
        '''Build a Dataset object from a JSON object.'''
        retval = Dataset(metadata=self._metadata)
        data = json_data[JSONKeys.DATA.value]
        if DatasetKeys.COVARIATES.value in data:
            covariates = pd.DataFrame(data[DatasetKeys.COVARIATES.value])
            retval[DatasetKeys.COVARIATES.value] = covariates
        if DatasetKeys.TARGET.value in data:
            target = pd.DataFrame(data[DatasetKeys.TARGET.value])
            retval[DatasetKeys.TARGET.value] = target
        if DatasetKeys.DATAFRAME.value in data:
            dataframe = pd.DataFrame(data[DatasetKeys.DATAFRAME.value])
            retval[DatasetKeys.DATAFRAME.value] = dataframe
        if DatasetKeys.GROUND_TRUTH.value in data:
            ground_truth = pd.DataFrame(data[DatasetKeys.GROUND_TRUTH.value])
            retval[DatasetKeys.GROUND_TRUTH.value] = ground_truth
        return retval



'''Compiles bound pipelines into executable pipelines'''
# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from pathlib import Path
from typing import Dict, Optional

from ..algorithms.impl.algorithm import AlgorithmCatalog
from ..executor.cucumber import JarOfCucumbers
from ..executor.executor_kind import ExecutorKind
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator, StepDesignator
from ..wrangler.saver import Saver
from .executable_pipeline import ExecutablePipeline, ExecutablePipelineStub


class InstantiatorError(Exception):
    '''An error occurred in use of the instantiator'''


class Instantiator(metaclass=abc.ABCMeta):
    '''Base class for pipeline instantiator kinds'''
    _kind: Optional[ExecutorKind] = None
    _saver: Optional[Saver] = None
    _algorithm_catalog: Optional[AlgorithmCatalog] = None

    def __init__(self,
                 saver: Optional[Saver] = None,
                 algorithm_catalog: Optional[AlgorithmCatalog] = None):
        self._saver = saver
        self._algorithm_catalog = algorithm_catalog

    @abc.abstractmethod
    def instantiate(self, pipeline: BoundPipeline) -> ExecutablePipeline:
        '''Instantiate a single bound pipeline into a single kind of executable pipeline.

        Optionally instantiate the pipeline with the given trained models.
        '''

    def instantiate_all(self,
                        pipelines: Dict[Designator, BoundPipeline]
                        ) -> Dict[Designator, ExecutablePipeline]:
        '''Instantiate a dict of bound pipelines into a dict of executable pipelines.'''
        retval: Dict[Designator, ExecutablePipeline] = {}

        for des, pipe in pipelines.items():
            retval[des] = self.instantiate(pipeline=pipe)

        return retval

    @abc.abstractmethod
    def save(self, pipeline: BoundPipeline, model_paths: Dict[StepDesignator, Path]) -> Path:
        '''Save a bound pipeline to disk using self._saver.

        We use model_paths to set the hyperparams, and to supply
        pretrained models if they are needed.

        Returns the path to that bound pipeline's saved file.
        '''


class InstantiatorStub(Instantiator):
    '''stub'''
    _kind = ExecutorKind('stub_executor_kind')
    _cucumbers: Optional[JarOfCucumbers]

    def __init__(self, cucumbers: Optional[JarOfCucumbers] = None, **kwargs):
        super().__init__(**kwargs)
        self._cucumbers = cucumbers

    def instantiate(self, pipeline: BoundPipeline) -> ExecutablePipeline:
        retval = ExecutablePipelineStub(
            kind=self._kind,
            pipeline=pipeline,
            cucumbers=self._cucumbers)
        return retval

    def save(self, pipeline: BoundPipeline, model_paths: Dict[StepDesignator, Path]) -> Path:
        return Path('stub')



'''A Table representing a numpy ndarray.'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List, Union

from io import BytesIO

import numpy as np
import pandas as pd

from .impl.table import TableFactory, Table, TableValueError
from .impl.table_catalog import TableCatalog


class NumpyTable(Table):
    '''A Table representing a numpy ndarray.'''
    name = 'numpy'
    _value: np.ndarray
    my_type = np.ndarray

    def to_csv(self, *args, **kwargs) -> str:
        '''Output the table as a CSV string.

        Extra parameters are passed to np.savetxt.
        '''
        # BUG: We do not emit column names, but numpy structured arrays (which have names)
        # are relatively niche.
        retval = BytesIO()
        np.savetxt(retval, self._value, *args, **kwargs)
        return retval.getvalue().decode()

    def as_dict(self, *args, **kwargs) -> Dict[Union[str, int], Any]:
        '''Convert table to a dictionary.'''
        return {name: self._value[i].tolist() for i, name in enumerate(self.columns)}

    def drop(self, columns: List[Union[int, str]]) -> Table:
        '''Return a copy of this table with selected columns dropped.'''
        if isinstance(columns[0], str):
            columns = [self.columns.index(c) for c in columns]
        return TableFactory(self._value[:, [i for i in range(self._value.shape[1])
                                            if i not in columns]])

    @property
    def empty(self) -> bool:
        '''Return True if the table is empty.

        For ndarrays, if any dimension is 0, it must be empty.
        '''
        return 0 in self._value.shape

    @property
    def columns(self) -> List[Union[int, str]]:
        '''Return the column names of the table.'''
        if self._value.dtype.names is not None:
            return self._value.dtype.names
        return list(str(i) for i, _ in enumerate(self._value, 1))

    @columns.setter
    def columns(self, value: List[Union[int, str]]) -> None:
        '''Set the column names of the table.'''
        raise TableValueError('Cannot set columns on a NumpyTable.')


def as_dataframe(from_table: Table) -> pd.DataFrame:
    '''Export a NumpyTable to a pandas DataFrame.'''
    assert isinstance(from_table, NumpyTable)
    return pd.DataFrame(data=from_table.value())


def register(catalog: TableCatalog):
    '''Register the NumpyTable with the catalog.'''
    catalog.register(NumpyTable)
    TableFactory.register_constructor(NumpyTable.my_type, NumpyTable)
    TableFactory.register_exporter(
        from_type=np.ndarray, to_type=pd.DataFrame, exporter=as_dataframe)
    TableFactory.register_exporter(
        from_type=np.ndarray, to_type=np.ndarray, exporter=NumpyTable.value)



""" Tests for dataframe_loader.py"""
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring, missing-class-docstring, duplicate-code

import copy
from typing import Dict, List, Optional, Union

import pandas as pd
import pytest

from ...config_components.dataset_config import DatasetConfig
from ...config_components.impl.config_component import ValidationErrors
from ...problem_def.problem_def import ProblemDefinition
from ...wrangler.dataset import Dataset, RoleName
from .dataframe_loader import ColumnError, DataframeLoader


CLAUSE = {
    'column_roles': {
        'target': {
            'name': 'a',
            'pos_label': 1
        }
    }
}


class FakeDataframeLoader(DataframeLoader):
    _train_df: Optional[pd.DataFrame] = None
    _test_df: Optional[pd.DataFrame] = None

    def __init__(self,
                 train_df: Optional[pd.DataFrame] = None,
                 test_df: Optional[pd.DataFrame] = None,
                 **kwargs):
        self._train_df = train_df
        self._test_df = test_df
        super().__init__(**kwargs)

    def _load_train(self) -> Dataset:
        if self._train_df is not None:
            return Dataset(dataframe=self._train_df)

        train_df = pd.DataFrame({
            'a': [1, 2],
            'b': [3, 4],
            'c': [5, 6],
            'd': [7, 8]})
        retval = Dataset()
        retval.dataframe = train_df
        return retval

    def _load_test(self) -> Optional[Dataset]:
        if self._test_df is None:
            return self._load_train()
        return Dataset(dataframe=self._test_df)

    def _poll(self, timeout: Optional[float] = 0.0) -> Optional[Dataset]:
        return Dataset(dataframe=self._train_df)


def test_attribute_role_default() -> None:
    config = DatasetConfig(clause=CLAUSE)
    dut = FakeDataframeLoader(config=config)
    dataset = dut.load_train()
    assert dataset is not None
    result = dataset.metadata.roles[RoleName.ATTRIBUTE]

    assert {n.name for n in result} == {'b', 'c', 'd'}


def test_no_target() -> None:
    '''Test that data with no target is acceptable'''
    clause_without_target = CLAUSE.copy()
    del clause_without_target['column_roles']
    config = DatasetConfig(clause=clause_without_target)
    dut = FakeDataframeLoader(config=config)

    dataset = dut.load_train()
    assert dataset is not None
    assert dataset.metadata.target is None


def test_validate_missing_col() -> None:
    '''Check for missing columns.

    If a role references a column that doesn't exist in the data,
    we want to throw a ValidationError.'''
    missing_col_clause = copy.deepcopy(CLAUSE)
    missing_col_clause['column_roles']['attribute'] = {
        'name': 'nonexistent_col'
    }
    config = DatasetConfig(clause=missing_col_clause)
    with pytest.raises(ValidationErrors, match='nonexistent_col'):
        FakeDataframeLoader(config=config)


def test_dataset_cols() -> None:
    config = DatasetConfig(clause=CLAUSE)
    dut = FakeDataframeLoader(config=config)
    got = dut.dataset(
        pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6]
        }),
        cols=['a']
    )

    want = pd.DataFrame({
        'a': [1, 2, 3]
    })

    pd.testing.assert_frame_equal(want, got.dataframe)


def test_dataset_roles() -> None:
    dut = FakeDataframeLoader(
        config=DatasetConfig({
            'column_roles': {
                'target': {'name': 'a'},
                'test_role': {'name': 'b'},
                'attribute': {'name': 'c'},
            }
        })
    )
    df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9],
        'd': [10, 11, 12]
    })
    got1 = dut.dataset(data=df, roles=[RoleName.TARGET])
    want1 = pd.DataFrame({
        'a': [1, 2, 3]})
    pd.testing.assert_frame_equal(got1.dataframe, want1)

    got2 = dut.dataset(data=df, roles=['attribute', RoleName.TEST_ROLE])
    want2 = pd.DataFrame({
        'b': [4, 5, 6],
        'c': [7, 8, 9],
        'd': [10, 11, 12],
    })
    pd.testing.assert_frame_equal(got2.dataframe, want2)


def test_dataset_roles_cols() -> None:
    dut = FakeDataframeLoader(
        config=DatasetConfig(clause={
            'column_roles': {
                'target': {'name': 'a'},
                'attribute': {'name': 'c'},
                'test_role': {'name': 'd'},
            }
        })
    )
    df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9],
        'd': [10, 12, 13]  # ROFL
    })
    got = dut.dataset(data=df, roles=['target'], cols=['c'])
    want = pd.DataFrame({
        'a': [1, 2, 3],
        'c': [7, 8, 9]
    })
    pd.testing.assert_frame_equal(got.dataframe, want)


def test_missing_columns() -> None:
    '''Because column d is present in the data but not assigned a role,
    it automatically gets assigned the attribute role.

    When we ask for a subset of the dataset containing all columns with
    the attribute role, we will try to select d, and we expect to raise an
    error because it is not present.
    '''
    dut = FakeDataframeLoader(
        config=DatasetConfig({
            'column_roles': {
                'target': {'name': 'a'},
                'test_role': {'name': 'b'},
            }
        })
    )
    df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9],
    })
    with pytest.raises(ColumnError, match="'d"):
        dut.dataset(data=df, roles=['attribute', RoleName.TEST_ROLE])


@pytest.mark.parametrize('argname,method,want_dict', [
    ('train_df', 'poll', {'a': [3], 'b': [6], 'c': [9]}),
    ('test_df', 'load_test', {'b': [4, 5, 6], 'c': [7, 8, 9]}),
    ('train_df', 'load_train', {'a': [3], 'b': [6], 'c': [9]}),
])
def test_split_dataset(argname: str, method: str, want_dict: Dict[str, List[int]]) -> None:
    '''Test that we split data properly in a distributed setting.

    Each node gets a subset of the data based on its node ID.
    '''
    pdef = ProblemDefinition({
        'dataset': {
            'column_roles': {
                'target': {'name': 'a'},
            }
        },
        'distributed': {
            'my_id': 3,
            'split': {
                'num_nodes': 3,
                'seed': 1234
            }
        },
        'problem_type': {
            'task': 'regression'
        }
    })
    config = pdef.get_conf('dataset')
    assert isinstance(config, DatasetConfig)
    df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9],
    })
    kwargs: Dict[str, Union[pd.DataFrame, DatasetConfig]] = {
        argname: df,
        'config': config
    }
    dut = FakeDataframeLoader(**kwargs)  # type: ignore[arg-type]

    want_df = pd.DataFrame(want_dict)
    # This is the test.
    got = getattr(dut, method)()
    assert got is not None
    pd.testing.assert_frame_equal(want_df, got.dataframe)


def test_build_dataset_from_json_predict() -> None:
    '''Test that we can build a dataset from a JSON object.'''
    pdef = ProblemDefinition({
        'dataset': {
            'column_roles': {
                'target': {'name': 'a'},
            }
        },
        'problem_type': {
            'task': 'regression'
        }
    })
    config = pdef.get_conf('dataset')
    assert isinstance(config, DatasetConfig)
    dut = FakeDataframeLoader(config=config)
    got = dut.build_dataset_from_json({'data': {'dataframe': {'a': [1, 2, 3], 'b': [4, 5, 6]}}})
    want = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
    })
    pd.testing.assert_frame_equal(want, got.dataframe)


def test_build_dataset_from_json_fit() -> None:
    '''Test that we can build a fit dataset from a JSON object.'''
    pdef = ProblemDefinition({
        'dataset': {
            'column_roles': {
                'target': {'name': 'a'},
            }
        },
        'problem_type': {
            'task': 'regression'
        }
    })

    request = {
        'data': {
            'covariates': {'b': [1, 2, 3], 'c': [4, 5, 6]},
            'target': {'a': [7, 8, 9]}
        },
    }
    config = pdef.get_conf('dataset')
    assert isinstance(config, DatasetConfig)
    dut = FakeDataframeLoader(config=config)
    got = dut.build_dataset_from_json(request)
    want_covariates = pd.DataFrame({
        'b': [1, 2, 3],
        'c': [4, 5, 6],
    })
    want_target = pd.DataFrame({
        'a': [7, 8, 9],
    })
    assert set(got.keys()) == {'covariates', 'target'}
    pd.testing.assert_frame_equal(got['covariates'], want_covariates)
    pd.testing.assert_frame_equal(got['target'], want_target)


def test_load_ground_truth() -> None:
    config = DatasetConfig(clause=CLAUSE)
    dut = FakeDataframeLoader(config=config)
    got = dut.load_ground_truth()
    assert got is not None
    pd.testing.assert_frame_equal(got.ground_truth,
                                  pd.DataFrame({'a': [1, 2]}))



'''IP Connection endpoint representing host and port number.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Tuple

from .endpoint import Endpoint


class IPEndpoint(Endpoint):
    '''IP Connection endpoint representing host and port number.'''
    _host: str
    _port: int

    def __init__(self, host: str, port: int):
        self._host = host
        self._port = port

    def __str__(self) -> str:
        return f'({self._host}, {self._port})'

    def __repr__(self) -> str:
        return f'({self._host}, {self._port})'

    def __hash__(self) -> int:
        return self.as_tuple.__hash__()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (self._host == other._host
                and self._port == other._port)

    @property
    def host(self) -> str:
        '''IP Address of this endpoint.'''
        return self._host

    @property
    def port(self) -> int:
        '''Port number of this endpoint'''
        return self._port

    @property
    def as_tuple(self) -> Tuple[str, int]:
        '''(host, port) as a tuple'''
        return (self._host, self._port)



'''Tests for json_loader.py'''
import json
from pathlib import Path
import pytest

from ..algorithms.impl.algorithm import AlgorithmCatalog, AlgorithmCatalogStub
from ..algorithms.impl.fake_algorithm import FakeAlgorithm, FakeInstance, FAKE_SERIALIZED
from ..algorithms.connect import ConnectorModel, ConnectorModelInstance
from ..executor.cucumber import Cucumber
from ..executor.simple.simple_executable_step import SimpleExecutableStep
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator, StepDesignator
from ..templates.impl.parallel_step import ParallelStep
from ..wrangler.dataset import Dataset
from .json_loader import JsonLoader


# pylint: disable = missing-function-docstring,protected-access,redefined-outer-name,duplicate-code


@pytest.fixture(scope='session')
def algorithm_catalog() -> AlgorithmCatalog:
    retval = AlgorithmCatalogStub()
    retval.register(FakeAlgorithm())
    retval.register(ConnectorModel())
    return retval


@pytest.fixture(scope='session')
def pipeline_file(tmp_path_factory: pytest.TempPathFactory) -> Path:
    output_path = tmp_path_factory.mktemp('output')
    contents = {
        'version': '1.0',
        'pipeline_designator': 'fake_bound',
        'pipeline_template_name': 'fake_bound',
        'output_dir': str(output_path),
        'steps': [
            {
                'type': 'parallel',
                'subpipelines': {
                    'left': [
                        {
                            'algorithm_name': 'fake_algorithm',
                            'model_filename': '@fake_algorithm_1@',
                            'type': 'step',
                            'queried': True
                        },
                    ],
                    'right': [
                        {
                            'algorithm_name': 'fake_algorithm',
                            'model_filename': '@fake_algorithm_3@named_step',
                            'type': 'step',
                            'name': 'named_step'
                        },
                    ]
                },
            },
            {
                'type': 'step',
                'algorithm_name': 'connect',
                'hyperparams': {
                    'newkey': ['left', 'oldkey'],
                    'othernewkey': ['right', 'oldkey']
                },
                'model_filename': '@connect_5@',
            }
        ],
    }
    pipeline_output_path = output_path / 'pipelines' / 'fake_bound.json'
    pipeline_output_path.parent.mkdir(parents=True, exist_ok=True)
    with open(pipeline_output_path, 'w', encoding='utf-8') as filepointer:
        json.dump(contents, filepointer)

    fake_inst = FakeInstance(parent=FakeAlgorithm())
    fake_inst.fit(dataset=Dataset())
    fake_pickle = Cucumber(
        impl=fake_inst,
        filename=StepDesignator('fake_step_des'),
        pipeline_designator=Designator('fake_bound'),
        hyperparams={}
    ).serialize()
    connector_pickle = Cucumber(
        impl=ConnectorModelInstance(parent=ConnectorModel()),
        filename=StepDesignator('@connect_5@'),
        pipeline_designator=Designator('fake_bound'),
        hyperparams={
            'newkey': ['left', 'oldkey'],
            'othernewkey': ['right', 'oldkey']
        }
    ).serialize()

    model_dir = output_path / 'models' / 'fake_bound'
    model_dir.mkdir(parents=True, exist_ok=True)

    with open(model_dir / '@fake_algorithm_1@.pkl', 'wb') as filepointer:
        filepointer.write(fake_pickle)

    with open(model_dir / '@fake_algorithm_3@named_step.pkl', 'wb') as filepointer:
        filepointer.write(fake_pickle)

    with open(model_dir / '@connect_5@.pkl', 'wb') as filepointer:
        filepointer.write(connector_pickle)

    # save a model to disk at output_folder/models/fake_bound/@connect_5@.pkl

    return pipeline_output_path


def test_load_pipeline_sunny_day(algorithm_catalog: AlgorithmCatalog, pipeline_file: Path) -> None:
    dut = JsonLoader(
        saver_version='1.0',
        algorithm_catalog=algorithm_catalog,
        pipeline_file=pipeline_file,
        load_models=True)

    got = dut.pipeline

    assert len(got.steps) == 2
    parallel_step = got.steps[0]
    connector_step = got.steps[1]
    assert isinstance(parallel_step, ParallelStep)
    assert connector_step.hyperparams() == {
        'newkey': ['left', 'oldkey'],
        'othernewkey': ['right', 'oldkey'],
    }
    left_pipeline = parallel_step.subpipelines['left']
    assert isinstance(left_pipeline, BoundPipeline)
    left_step = left_pipeline.steps[0]
    # Instantiate left_step as kind 'simple'.
    left_executable = SimpleExecutableStep(left_step)
    left_instance = left_executable._model_instance
    assert isinstance(left_instance, FakeInstance)
    assert left_instance.evidence_of_deserialize == FAKE_SERIALIZED



'''Tests for SimpleParallelStep'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...algorithms.connect import ConnectorModel
from ...generator.bound_pipeline import BoundPipeline
from ...wrangler.dataset import Dataset

from .simple_parallel_executable_step import SimpleParallelExecutableStep
from .simple_executable_pipeline import SimpleExecutablePipeline

# pylint: disable=missing-function-docstring,duplicate-code

DATASET = Dataset(
    a='a_value',
    b='b_value'
)


def test_sunny_day() -> None:
    pipeline = BoundPipeline(name='Fake_bound_pipeline')
    pipe1 = pipeline.new(name='at')
    pipe1.step(ConnectorModel(c='a')).set_name('connector_pipe1')
    pipe2 = pipeline.new(name='dt')
    pipe2.step(ConnectorModel(d='b')).set_name('connector_pipe2')
    step = pipeline.parallel(left=pipe1, right=pipe2).set_name('parallel_step')

    dut = SimpleParallelExecutableStep(step, SimpleExecutablePipeline)

    dut.fit(dataset=DATASET)
    got = dut.predict(dataset=DATASET)
    assert got == Dataset(
        left=Dataset(c='a_value'),
        right=Dataset(d='b_value')
    )



'''Tests for pipeline_result.py'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd
import pytest

from ..generator.bound_pipeline import BoundPipelineStub
from ..generator.designator import Designator
from ..splitters.impl.splitter import SplitDataset, Fold
from ..wrangler.dataset import Dataset, Metadata, RoleName, Column
from .executable_pipeline import ExecutablePipelineStub
from .executable_pipeline import PipelineResult, PipelineResultError, PipelineResults

# pylint: disable=missing-function-docstring,duplicate-code


def test_executable_pipelines_sunny_day() -> None:
    pr1 = PipelineResult(
        prediction=Dataset(),
        executable_pipeline=ExecutablePipelineStub(
            pipeline=BoundPipelineStub(name='stub pipeline 1'),
            trained=True))
    pr2 = PipelineResult(
        prediction=Dataset(),
        executable_pipeline=ExecutablePipelineStub(
            pipeline=BoundPipelineStub(name='stub pipeline 2'),
            trained=True))
    des1 = Designator('des1')
    des2 = Designator('des2')
    dut = PipelineResults({
        des1: pr1, des2: pr2})
    assert des1 in dut.executable_pipelines
    assert des2 in dut.executable_pipelines
    assert dut.executable_pipelines[des1].name == 'stub pipeline 1'
    assert dut.executable_pipelines[des2].name == 'stub pipeline 2'


def test_executable_pipelines_fail() -> None:
    '''Test that PipelineResult properly raises an error
    if executable_pipelines property is called when not all
    results have executable pipelines.
    '''
    pr1 = PipelineResult(
        prediction=Dataset(),
        executable_pipeline=ExecutablePipelineStub(
            pipeline=BoundPipelineStub(name='stub pipeline 1'),
            trained=True))
    bound_only = PipelineResult(
        prediction=Dataset(),
        bound_pipeline=BoundPipelineStub(
            name='bound stub pipeline'))
    des1 = Designator('des1')
    des2 = Designator('des2')
    dut = PipelineResults({
        des1: pr1, des2: bound_only})

    with pytest.raises(PipelineResultError, match='bound stub pipeline'):
        _ = dut.executable_pipelines


def test_add_ground_truth() -> None:
    empty = Dataset(
        dataframe=pd.DataFrame(),
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('a')]}
        )
    )
    pred = empty.output()
    pred.predictions = pd.DataFrame(
        {'a': [1, 2, 3]})
    gt = empty.output()
    gt.ground_truth = pd.DataFrame(
        {'a': [4, 5, 6]}
    )
    dut = PipelineResult(
        prediction=pred,
        bound_pipeline=BoundPipelineStub('stub bound pipeline'),
        split_dataset=SplitDataset([
            Fold(
                train=empty,
                validate=empty,
                ground_truth=gt
            )
        ]))
    assert dut.prediction is not None
    got = dut.prediction.predictions
    want = pd.DataFrame({
        'a': [1, 2, 3],
        'ground truth': [4, 5, 6]
    })
    pd.testing.assert_frame_equal(got, want)



'''Sends and recieves messages using queues in memory.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from collections import defaultdict
from enum import Enum
from queue import Queue
from typing import Any, Dict, List, Optional, Set

from aenum import Enum as AEnum  # type: ignore[import-untyped]


from ..config_components.impl.config_component import ConfigComponent
from ..config_components.distributed_config import CommunicatorConfig, DistributedConfig
from ..neighbor_manager.event import NewNeighbor, Recv
from ..neighbor_manager.node_id import NodeID
from ..wrangler.constants import Defaults
from ..wrangler.logger import Level, Logger

from .impl.communicator import Communicator
from .impl.communicator_catalog import CommunicatorCatalog

logger = Logger(__file__, level=Level.INFO).logger()


class MemoryConfig(ConfigComponent):
    '''Specifies Memory configuration in a distributed AI setting.'''
    name = 'memory'
    tags: Dict[str, Any] = {}
    _my_id: NodeID

    def __init__(self, clause: Dict[str, Any], my_id: NodeID, name: Optional[str] = None) -> None:
        super().__init__(clause)
        if name is not None:
            self.name = name
        self._my_id = my_id

    class Constants(Enum):
        '''Keys below the top level.'''

    class Keys(AEnum):
        '''Valid keys for the top level.'''
        DOMAIN = 'domain'

    class Defaults(Enum):
        '''Default values'''

    def required_keys(self) -> Set[str]:
        return {
            self.Keys.DOMAIN.value,  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        }

    @property
    def domain(self) -> str:
        '''Communication Domain for memory.

        All nodes in the same network need to pick the same domain.
        '''
        return self._get(self.Keys.DOMAIN)


class MemoryCommunicator(Communicator):
    '''Sends and recieves messages using UDP memory.'''
    name: str = 'memory'
    tags: Dict[str, List[str]] = {}

    # Class variables
    # Neighbors should introduce themselves to us before communicating.
    # First key is domain, second key is node ID.
    _introductions: Dict[str, Dict[NodeID, Set[NodeID]]] = defaultdict(dict)
    # These are the queues that the receiving nodes will listen to.
    # First key is domain, second key is node ID.
    _queues: Dict[str, Dict[NodeID, Queue]] = defaultdict(dict)

    _domain: str = 'default'
    _timeout: Optional[float] = None

    def __init__(self,
                 known_neighbors: Set[NodeID],
                 my_id: NodeID,
                 distributed: DistributedConfig):
        super().__init__(my_id=my_id, known_neighbors=known_neighbors, distributed=distributed)
        self._domain = self._distributed.communicator.memory.domain  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        self._introductions[self._domain][self._my_id] = known_neighbors

    def stop(self):
        '''Purge our queue.'''
        del self._queues[self._domain][self._my_id]
        if not self._queues[self._domain]:
            del self._queues[self._domain]
        self._timeout = None

    def start(self,
              queue: Queue,
              timeout: float = Defaults.LISTENER_TIMEOUT) -> None:
        '''Start the communicator.'''
        self._timeout = timeout
        self._queues[self._domain][self._my_id] = queue

    def send(self,
             dest_ids: List[NodeID],
             payload: bytes) -> int:
        '''Send a message to the nodes in dest_ids.

        Return value is the size of the messages sent.
        '''
        retval = 0
        assert self._timeout is not None, 'Communicator not started'
        for dest_id in dest_ids:
            try:
                d_queue = self._queues[self._domain][dest_id]
                d_introductions = self._introductions[self._domain][dest_id]
            except KeyError:
                logger.error('Unknown destination ID: %s. Has desination been started?', dest_id)
                continue
            retval += len(payload)
            if self._my_id not in d_introductions:
                d_introductions.add(self._my_id)
                d_queue.put(NewNeighbor(neighbor=self._my_id), timeout=self._timeout)
            d_queue.put(Recv(self._my_id, payload), timeout=self._timeout)

        return retval

    def send_all(self, payload: bytes) -> int:
        '''Send a message to all started neighbors.

        Return value is a count of neighbors we sent to.
        '''
        dest_ids = [dest_id
                    for dest_id in self._queues[self._domain].keys()
                    if dest_id != self._my_id]
        self.send(dest_ids=dest_ids, payload=payload)
        return len(dest_ids)


def register(catalog: CommunicatorCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(MemoryCommunicator)
    CommunicatorConfig.register(name="memory", config_type=MemoryConfig)



'''Tests for problem_def.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path
import subprocess

import pandas as pd
import pytest

from ..config_components.dataset_config import DatasetConfig
from ..config_components.impl.config_component import ParsingErrors
from .output_config import OutputConfigError
from .problem_def import ProblemDefinition, Task, ValidationErrors

# pylint: disable=missing-function-docstring,protected-access
# pylint: disable=redefined-outer-name,duplicate-code


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    return tmp_path_factory.mktemp("data")


def module_path() -> Path:
    pathobj = Path(__file__).resolve()
    module_parent = pathobj.parents[2]
    return module_parent


TEST_PROBLEM_DEF = '''{
    "_comments" : [
        "A json file fully encapsulating the problem definition",
        "Specifically, an exmaple regression problem."
    ],
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "output": {}
}
'''


def test_get():
    dut = ProblemDefinition(TEST_PROBLEM_DEF)
    assert dut._get('problem_type', 'task') == 'regression'
    assert set(dut._get().keys()) == set(
        ['output', 'dataset', 'problem_type'])


def test_read_from_file(tmp_path: Path) -> None:
    tmp_file = tmp_path / 'pdef.json'
    with open(str(tmp_file), 'w', encoding='utf8') as f:
        f.write(TEST_PROBLEM_DEF)
    dut = ProblemDefinition(str(tmp_file))
    assert dut._get('problem_type', 'task') == 'regression'
    assert set(dut._get().keys()) == set(
        ['output', 'dataset', 'problem_type'])


def test_read_from_file_fail() -> None:
    '''We want to throw multiple errors for an invalid string.

    If a string is given that cannot be parsed as JSON or a filepath,
    we want to show the errors from both attempts to parse it.
    '''
    match_str = r'(JSONDecode.*FileNotFound)|(FileNotFound.*JSONDecode)'
    with pytest.raises(ParsingErrors,
                       match=match_str):
        ProblemDefinition('foo.foo')


def test_dataset_config() -> None:
    path = module_path() / 'examples' / 'classification' / 'credit.csv'
    pdef = {
        "dataset": {
            "config": "local",
            "train_path": path,
            "column_roles": {}
        },
        "problem_type": {
            "task": "regression"
        },
        "output": {}
    }
    dut = ProblemDefinition(pdef)
    dataset_config = dut.get_conf(dut.Keys.DATASET)
    assert isinstance(dataset_config, DatasetConfig)
    assert 'local' == dataset_config.config


MISSING_KEY = '''{
    "metrics" : {}
}
'''

# Missing "output" key


def test_missing_keys() -> None:
    with pytest.raises(ValidationErrors, match=r"[Mm]issing"):
        ProblemDefinition(MISSING_KEY)


INVALID_KEY = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "hamster": {}
}
'''


def test_invalid_keys() -> None:
    with pytest.raises(ValidationErrors, match=r"hamster"):
        ProblemDefinition(INVALID_KEY)


METRIC_MULTI = '''{
    "metrics" : {
        "metric1" : {
            "catalog_name": "real_metric_name",
            "k1" : "v1"
        },
        "metric2" : {
            "catalog_name": "real_metric_name",
            "k1" : "v2"
        },
        "another_metric": {}
    },
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    }
}
'''


def test_metric_sunny_day() -> None:
    dut = ProblemDefinition(METRIC_MULTI).metric_configs
    assert len(dut) == 3
    dutm1 = dut['metric1']
    assert dutm1.name == 'metric1'
    assert dutm1.hyperparams == {'k1': 'v1'}
    assert dutm1.catalog_name == 'real_metric_name'
    dutm2 = dut['metric2']
    assert dutm2.name == 'metric2'
    assert dutm2.hyperparams == {'k1': 'v2'}
    assert dutm2.catalog_name == 'real_metric_name'
    dutanother = dut['another_metric']
    assert dutanother.name == 'another_metric'
    assert dutanother.hyperparams == {}
    assert dutanother.catalog_name == 'another_metric'


def test_no_metrics() -> None:
    dut = ProblemDefinition(TEST_PROBLEM_DEF)
    assert dut.metric_configs == {}


SUNNY_TASK = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "data_type": "text",
        "task": "regression"
    }
}'''


def test_task_sunny_day() -> None:
    dut = ProblemDefinition(SUNNY_TASK)
    prob_type = dut.task
    print(f'Problem type: {prob_type}')
    assert prob_type == Task(clause={'data_type': 'text', 'task': 'regression'})


TASK_WITH_NO_TASK_TYPE = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "data_type": "TABULAR"
    }
}'''


def test_problem_type_missing_task() -> None:
    with pytest.raises(ValidationErrors, match=r'[Tt]ask'):
        ProblemDefinition(TASK_WITH_NO_TASK_TYPE)


TASK_WITH_NO_DATA_TYPE = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "REGRESSION"
    }
}'''


def test_problem_type_missing_data_type() -> None:
    dut = ProblemDefinition(TASK_WITH_NO_DATA_TYPE)
    want = Task(clause={'data_type': "tabular", 'task': "regression"})
    got = dut.task
    assert want == got


NON_DICT_PROBLEM_TYPE = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": []
}'''


def test_problem_type_not_dict() -> None:
    with pytest.raises(ValidationErrors,
                       match=r'(?i)problem_type.*dict'):
        ProblemDefinition(NON_DICT_PROBLEM_TYPE)


_DEFAULT_INSTANTIATIONS = sorted(['simple', 'stub_executor_kind'])


def test_instantiations_missing() -> None:
    dut = ProblemDefinition(TEST_PROBLEM_DEF)
    instantiations = dut.output.instantiations
    assert sorted(instantiations) == _DEFAULT_INSTANTIATIONS


EMPTY_INSTANTIATIONS = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "output": {
        "instantiations": []
    }
}'''


def test_instantiations_disable() -> None:
    dut = ProblemDefinition(EMPTY_INSTANTIATIONS)
    instantiations = dut.output.instantiations
    assert not instantiations


SIMPLE_INSTANTIATIONS = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "task": "regression"
    },
    "output": {
        "instantiations": [
            "SIMPLE",
            "stub_executor_kind"
        ]
    }
}'''


def test_instantiations_sunny_day() -> None:
    dut = ProblemDefinition(SIMPLE_INSTANTIATIONS)
    instantiations = dut.output.instantiations
    assert set(instantiations) == set({'simple', 'stub_executor_kind'})


COMMENTS_PRESENT = '''{
    "metrics" : {},
    "dataset": {
        "config": "ignore",
        "column_roles": {}
    },
    "problem_type": {
        "_comments": ["Some comment."],
        "data_type": "TEXT",
        "task": "REGRESSION"
    }
}'''


def test_comments() -> None:
    dut = ProblemDefinition(COMMENTS_PRESENT)
    got = dut._get('problem_type')
    want = {
        'data_type': 'TEXT',
        'task': 'REGRESSION'
    }
    assert got == want


BINARY_CLASSIFICATION_NO_POS_LABEL = '''{
    "dataset": {
        "config": "ignore",
        "column_roles": {
            "target": "class"
        }
    },
    "problem_type": {
        "task": "binary_classification"
    },
    "metrics" : [
        {
          "MEAN_ABSOLUTE_ERROR": {}
        }
    ]
}
'''


@pytest.mark.xfail
def test_binary_classification_no_pos_label() -> None:
    with pytest.raises(ValidationErrors, match='binary_classification.*pos_label'):
        ProblemDefinition(BINARY_CLASSIFICATION_NO_POS_LABEL)


OUTPUT_DIR = '''{{
    "dataset": {{
        "config": "ignore",
        "column_roles": {{}}
    }},
    "problem_type": {{
        "task": "regression"
    }},
    "output": {{
        "path": "{output_path}"
    }}
}}
'''


def test_output_dir_exists(tmp_path: Path) -> None:
    tmp_dir = tmp_path / "sub"
    tmp_dir.mkdir()
    pdef_json = OUTPUT_DIR.format(output_path=str(tmp_dir))
    dut = ProblemDefinition(pdef_json)
    assert dut.output.path == tmp_dir


def test_output_dir_not_exist(tmp_path: Path) -> None:
    newsubdir = tmp_path / "newsubdir"
    pdef_json = OUTPUT_DIR.format(output_path=str(newsubdir))
    dut = ProblemDefinition(pdef_json)
    assert dut.output.path == newsubdir
    assert newsubdir.exists()


def test_output_dir_not_a_dir(tmp_path: Path) -> None:
    notadir = tmp_path / "notadir.txt"
    notadir.write_text('This is a text file, not a directory.\n')
    pdef_json = OUTPUT_DIR.format(output_path=str(notadir))
    with pytest.raises(OutputConfigError):
        _ = ProblemDefinition(pdef_json)


def test_output_dir_parent_not_exist(tmp_path: Path) -> None:
    nonexistent_parent = tmp_path / 'nonexistent' / 'also_nonexistent'
    pdef_json = OUTPUT_DIR.format(output_path=str(nonexistent_parent))
    with pytest.raises(OutputConfigError):
        _ = ProblemDefinition(pdef_json)


MEMORY_PROBLEM_DEF_WITH_K = '''{
    "dataset": {
        "config": "memory",
        "column_roles": {
            "target": {
                "name": "c"
            }
        },
        "params": {
            "train_data": "dataframe"
        }
    },
    "problem_type": {
        "task": "binary_classification"
    },
    "cross_validation": {
        "k": 10
    }
}
'''


def test_memory_problem_def_with_k() -> None:
    dataframe = pd.DataFrame(  # noqa: F841 pylint: disable=unused-variable
        {
            'a': range(1, 15001),
            'b': range(1, 15001),
            'c': range(1, 15001)
        }
    )
    dut = ProblemDefinition(MEMORY_PROBLEM_DEF_WITH_K)

    assert dut.cross_validation_config.k == 10


# TODO(piggy): This test relies on validation that was removed to make image processing work.
# The fix is to add a new catalog for problem defs.
@pytest.mark.xfail
def test_memory_problem_def_with_invalid_k() -> None:
    dataframe = pd.DataFrame(  # noqa: F841 pylint: disable=unused-variable
        {
            'a': range(1, 3),
            'b': range(1, 3),
            'c': range(1, 3)
        }
    )

    with pytest.raises(ValidationErrors, match=r'(k.*row)|(row.*k)/i'):
        ProblemDefinition(MEMORY_PROBLEM_DEF_WITH_K)


def test_plugin_config_component() -> None:
    '''Test that plugins can add config components.

    There is a race with other tests that load the problem_def catalog.
    Between installing and removing the plugin, the catalog can be partially
    loaded by another test, which then finishes loading after the removal.
    This triggers a ModuleNotFoundError. This happens something like 3% of
    the time. It will not happen in CI on gitlab because those tests are
    single-threaded.
    '''

    pluginpath = (Path(__file__).parents[2] / 'plugins' / 'testproblemdef'
                  / 'dist' / 'testproblemdef-0.0.1-py3-none-any.whl')
    subprocess.run(['python', '-m', 'build'], cwd=pluginpath.parents[1], check=True)
    subprocess.run(['pip', 'install', pluginpath], check=False)
    try:
        ProblemDefinition({
            'dataset': {
                'config': 'ignore'
            },
            'problem_type': {
                'task': 'test_task'
            },
            'widget_config': {
                'sprocket': 'sprocket_value'
            }
        })
    finally:
        subprocess.run(['pip', 'uninstall', '-y', 'testproblemdef'], check=False)



'''These are the kinds of executable pipelines for different executors.'''
from typing import Set

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


# TODO(piggy): If we introduce an ExecutorCatalog, the constructor for
# ExecutorKind should confirm that the kind exists in the catalog.
class ExecutorKind(str):
    '''The class for the Executor Kind identifying the type of Executor'''
    _ALLOWED = {'json', 'stub_executor_kind', 'simple'}

    @staticmethod
    def __new__(cls, content):
        if content.lower() not in cls._ALLOWED:
            raise NotImplementedError(
                f'Executor kind {content} is not recognized. '
                f'Existing executor kinds: {cls._ALLOWED}'
            )
        return str.__new__(cls, content.lower())

    @classmethod
    def allowed(cls) -> Set[str]:
        '''Currently supported executor kinds'''
        return cls._ALLOWED

    @property
    def suffix(self) -> str:
        '''Filename suffix used for saving pipelines to disk'''
        if self == 'json':
            return '.json'
        if self == 'stub_executor_kind':
            return '.stub'
        raise NotImplementedError(
            f'No file suffix available for ExecutorKind "{self}".')



'''tests for cross_validator.py'''
# pylint: disable=missing-function-docstring,missing-class-docstring
# pylint: disable=protected-access,duplicate-code

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..stub_cross_validator import StubCrossValidator


class FakeCV(StubCrossValidator):
    _name = 'fake_cv_default_hyperparams'
    _hyperparams = {
        'k1': 'v1',
        'k2': 'v2'
    }


def test_hyperparam_overrides() -> None:
    dut = FakeCV()
    assert dut._hyperparams == {
        'k1': 'v1',
        'k2': 'v2'
    }
    assert dut.hyperparams(k2='new_val', new_key='new_val2') == {
        'k1': 'v1',
        'k2': 'new_val',
        'new_key': 'new_val2'
    }


class FakeCVNoDefaults(StubCrossValidator):
    _name = 'fake_cv_no_default_hyperparams'


def test_no_default_hyperparams() -> None:
    dut = FakeCVNoDefaults()
    assert not dut._hyperparams



'''Save results.'''
from pathlib import Path

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.
from typing import Dict, Optional
import pandas as pd

from ..executor.executor_kind import ExecutorKind
from ..generator.designator import Designator, StepDesignator
from ..instantiator.executable_pipeline import ExecutablePipeline
from ..instantiator.executable_pipeline import PipelineResult, PipelineResults
from ..problem_def.output_config import OutputConfig

from .constants import Filename, FileType, OutputFolder
from .dataset import Dataset


def save_csv(path: Path, filename: Filename, dataset: Optional[Dataset]) -> None:
    '''Save the dataset as a csv at the specified path'''
    if dataset is None:
        raise NotImplementedError('Cannot save a None dataset.')
    for key, val in dataset.items():
        outfile = path / (f'{filename.value}_{key}.{FileType.CSV.value}')
        if not isinstance(val, pd.DataFrame):
            raise NotImplementedError(f'Cannot save a {type(val)}.')
        val.to_csv(outfile, index=False)


def save_arff(path: Path, filename: Filename, dataset: Optional[Dataset]) -> None:
    '''Save dataset as an arff file.'''
    raise NotImplementedError()


Savers = {
    FileType.CSV: save_csv,
    FileType.ARFF: save_arff
}


class Saver():
    '''Save results from the wrangler.'''
    _output_config: OutputConfig
    _current_ver: str

    def __init__(self, output_config: OutputConfig):
        self._output_config = output_config
        self._current_ver = '1.0'

    @property
    def current_version(self) -> str:
        '''Current version of directory system for saving pipelines and models.

        Needed for backward compatibility of saved files.
        '''
        return self._current_ver

    @property
    def output_path(self) -> Path:
        '''The root of the output hierarchy.'''
        assert self._output_config.path is not None, (
            'BUG: Saver is only built if there is an output path.'
        )
        return self._output_config.path

    def save(self, result: PipelineResult) -> Path:
        '''Save train predictions, and optionally test predictions, for a single PipelineResult.

        Operates according to this saver's output config.
        Returns the Path to the folder containing the predictions.
        '''

        foldername = result.bound_pipeline.designator
        folderpath = self.output_path / foldername
        folderpath.mkdir(exist_ok=True)

        Savers[self._output_config.file_type](
            path=folderpath,
            filename=Filename.TRAIN_PREDICTIONS,
            dataset=result.prediction
        )

        return folderpath

    def save_all_models(self, results: PipelineResults) -> Dict[StepDesignator, Path]:
        '''Save all the models in a PipelineResults.

        Returns a Path object for each saved step relative to the output root.
        '''
        return self.save_models(results.executable_pipelines)

    def model_path(self, des: Designator, model_filename: str) -> Path:
        '''The absolute path to a saved model, given its filename and pipeline designator.'''
        if self.current_version == '1.0':
            return self.output_path / OutputFolder.MODELS / des / model_filename
        raise NotImplementedError('model_path only knows version 1.0')

    def save_models(self,
                    pipelines: Dict[Designator, ExecutablePipeline]) -> Dict[StepDesignator, Path]:
        '''Save all the models in every executable pipeline.

        Returns a Path object for each saved step relative to the output root.
        '''
        retval = {}
        for pipedesignator, pipeline in pipelines.items():
            for stepdesignator, cucumber in pipeline.cucumberize_all().items():
                filename = cucumber.filename + '.pkl'
                model_path = self.model_path(des=pipedesignator, model_filename=filename)
                model_path.parent.mkdir(parents=True, exist_ok=True)
                with open(model_path, 'wb') as file_pointer:
                    file_pointer.write(cucumber.serialize())
                retval[stepdesignator] = Path(filename)
        return retval

    def save_pipeline(self, pipeline: bytes, des: Designator, kind: ExecutorKind) -> Path:
        '''Save a serialized pipeline in a file based on its designator and kind.'''
        pipepath = self.output_path / OutputFolder.PIPELINES
        pipepath.mkdir(exist_ok=True)
        filename = f'{des}{kind.suffix}'
        with open(pipepath / filename, 'wb') as file_pointer:
            file_pointer.write(pipeline)
        return pipepath / filename



'''Tests for numpy.py.'''
import numpy as np
import pandas as pd

from .impl.table_auto import TableCatalogAuto
from .impl.table import TableFactory

from .numpy import NumpyTable


def test_numpy_table():
    '''Test the NumpyTable.'''
    _ = TableCatalogAuto()

    my_ndarray = np.array([[1, 2, 3, 7], [4, 5, 6, 8]])
    my_table = TableFactory(my_ndarray)  # pylint: disable=abstract-class-instantiated

    # assert isinstance(my_table, Table)
    assert isinstance(my_table, NumpyTable)

    np.testing.assert_array_equal(my_table.value(), my_ndarray)

    assert my_table[0].tolist() == [1, 2, 3, 7]
    assert my_table[1].tolist() == [4, 5, 6, 8]

    my_table[0] = my_table[0] + my_table[1]
    assert my_table[0].tolist() == [5, 7, 9, 15]

    got_ndarray = my_table.as_(np.ndarray)
    assert isinstance(got_ndarray, np.ndarray)
    assert got_ndarray[0, 2] == 9
    assert got_ndarray.shape == (2, 4)

    got_df = my_table.as_(pd.DataFrame)
    assert isinstance(got_df, pd.DataFrame)
    assert got_df.iloc[0, 2] == 9
    assert got_df.shape == (2, 4)
    assert all(got_df.columns == [0, 1, 2, 3])



'''Catalog for metrics that can be used to evaulate pipelines'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, List, Optional, Type, Union

from ...catalog.memory_catalog import MemoryCatalog
from .config_component import ConfigComponent


class ConfigComponentCatalog(MemoryCatalog[Type[ConfigComponent]], metaclass=abc.ABCMeta):
    '''Base class for config component catalogs.'''

    def register(self, obj: Type[ConfigComponent], name: Optional[str] = None,
                 tags: Optional[Dict[str, Union[str, List[str]]]] = None) -> str:
        return super().register(obj, name, tags)


class ConfigComponentCatalogStub(ConfigComponentCatalog):
    '''stub'''



'''tests for distributed_config.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code

from pathlib import Path
import pytest

from ..neighbor_manager.node_id import NodeID

from .impl.config_component import ValidationErrors
from .distributed_config import DistributedConfig


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("distributed_config", numbered=True)
    return retval


def test_get_static_adjacency() -> None:
    dut = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2, 4, 5],
                    '2': [1, 3, 5],
                    '3': [5, 2],
                    '4': [1],
                    '5': [1, 2, 3]
                }
            }
        }
    })

    static_adjacency = dut.get_static_adjacency(my_id=NodeID(3))
    assert static_adjacency is not None
    assert set(static_adjacency) == {NodeID(2), NodeID(5)}


def test_get_static_from_edges() -> None:
    dut = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'edges': [
                    [1, 2],
                    [1, 4],
                    [1, 5],
                    [2, 3],
                    [2, 5],
                    [3, 5]
                ]
            }
        }
    })

    assert dut.get_static_adjacency(my_id=NodeID(3)) == [NodeID(2), NodeID(5)]


def test_validation_edges_wrong_length() -> None:
    dut = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'edges': [
                    [1, 2],
                    [1, 4],
                    [1, 5],
                    [2, 3],
                    [2, 5],
                    [3, 5, 7]
                ]
            }
        }
    })
    with pytest.raises(ValidationErrors, match=r'\[3, 5, 7\]'):
        dut.validate()


def test_baseline_djam() -> None:
    dut = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '0': [1, 2],
                    '1': [0],
                    '2': [0]
                }
            }
        },
        'regularization': {
            'name': 'djam',
            'weight_matrix': [
                [0, .5, .7],
                [.5, 0, 0],
                [.7, 0, 0]
            ]
        },
        'my_id': 1
    })

    assert dut.regularization_type == 'djam'
    assert isinstance(dut.neighbor_weights, dict)
    assert dut.neighbor_weights[NodeID(0)] == 0.5
    assert dut.neighbor_weights[NodeID(1)] == 0.0


def test_dropper(tmp_path) -> None:
    dut = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [],
                }
            }
        },
        'communicator': {
            'name': 'stub_communicator',
        },
        'dropper': {
            'drop_rate': 0.9,
            'seed': 42,
            'output_dir': tmp_path,
        },
        'my_id': 1,
    })

    assert dut.dropper is not None
    assert dut.dropper.drop_rate == 0.9
    assert dut.dropper.seed == 42
    assert dut.dropper.output_dir == tmp_path


def test_unicast_kafka() -> None:
    dut = DistributedConfig(clause={
        'communicator': {
            'name': 'kafka',
            'kafka': {
                'bootstrap_servers': 'localhost:9092',
                'topic': 'test',
                'group_id': 'test',
                'topic_prefix': 'prefix_',
            },
            'strategy': 'unicast',
        },
        'my_id': 1,
    })

    assert dut.communicator.strategy == 'unicast'
    assert dut.communicator.kafka.node_topic() == 'prefix_1'  # type: ignore[attr-defined] # pylint: disable=no-member



'''A fake distributed algorithm used for tests.'''

from typing import Any, Optional

import pickle

from ...config_components.distributed_config import DistributedConfig
from ...wrangler.dataset import Dataset
from ...wrangler.logger import Logger, Level

from .algorithm import Algorithm
from .distributed_algorithm_instance import (
    DistributedAlgorithmInstance, NeighborState)

log = Logger(__file__, to_stdout=False, level=Level.DEBUG).logger()


class FakeDistributedAlgorithm(Algorithm):
    '''A fake distributed algorithm used for tests.'''
    _name = 'fake_distributed_algorithm'
    _default_hyperparams = {  # pylint: disable=duplicate-code
        'Lambda': 1.0,
        'omega': 0.667,
        'tol': 0.000000001,
        'maxiter': None,
    }

    def instantiate(self, **hyperparams) -> 'FakeDistributedInstance':
        return FakeDistributedInstance(parent=self, **hyperparams)


class FakeDistributedAlgorithmNeighbor(NeighborState):
    '''A fake distributed algorithm neighbor state.'''
    def encode(self) -> bytes:
        '''Encode a message for distributed neighbors. '''
        return pickle.dumps(self._columns)

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'FakeDistributedAlgorithmNeighbor':
        '''Decode a message from distributed neighbors.'''
        columns = pickle.loads(serialized_model)
        return cls(columns=columns)

    def state_differs(self,
                      distributed: DistributedConfig,
                      other: Optional[Any]) -> bool:
        '''Decide whether to send a message to neighbors.

        last_state_sent is an instance of type(self) or None.
        '''
        _ = distributed
        return other is None  # Only send once

    def distance(self, other: 'FakeDistributedAlgorithmNeighbor') -> float:
        return 0.0


FAKE_SERIALIZED = b'fake serialized algorithm'


class FakeDistributedInstance(DistributedAlgorithmInstance):
    '''An instance of FakeAlgorithm for tests.'''
    evidence_of_deserialize = None
    _neighbor_constructor = FakeDistributedAlgorithmNeighbor

    def __init__(self, parent, distributed, **kwargs):
        self.hyperparams = parent.hyperparams(**kwargs)
        super().__init__(parent=parent, distributed=distributed, **parent.hyperparams(**kwargs))

    def deserialize(self, serialized_model: bytes) -> 'FakeDistributedInstance':
        self.evidence_of_deserialize = serialized_model
        return self

    def serialize(self) -> bytes:
        return FAKE_SERIALIZED

    def _decode(self, serialized_model: bytes) -> FakeDistributedAlgorithmNeighbor:
        '''Decode a message from distributed neighbors. '''
        return FakeDistributedAlgorithmNeighbor.decode(serialized_model)

    def _fit(self, dataset: Optional[Dataset], **kwargs) -> None:
        '''If we get data, update our state to match its columns.

        If we get no data but have seen a neighbor, update our state to match the
            neighbor's columns.
        '''
        if dataset is not None:
            self._my_state = FakeDistributedAlgorithmNeighbor(
                columns=list(dataset.dataframe.columns))
        elif self._neighbor_metadata:
            self._my_state = FakeDistributedAlgorithmNeighbor(
                columns=next(self._neighbor_models_iter).columns)  # type: ignore
            # not sure why mypy complains about using next() as above,
            #   we use it identically in other places.

    def _predict(self, dataset: Optional[Dataset], **kwargs) -> Optional[Dataset]:
        '''Apply model to input dataset to create output.

        This handles model-speific work and does not handle general tasks.
        '''
        return dataset

    @property
    def my_state(self) -> Optional[FakeDistributedAlgorithmNeighbor]:
        '''Accessor so that tests can see the state'''
        retval = self._my_state
        if retval is not None:
            assert isinstance(retval, FakeDistributedAlgorithmNeighbor)
        return retval


class FitCountDistributedAlgorithm(Algorithm):
    '''A fake distributed algorithm that counts the number of times it has been fit'''
    _name = 'fit_count_distributed_algorithm'
    _default_hyperparams = {
        'Lambda': 1.0,
        'omega': 0.667,
        'tol': 0.000000001,
        'maxiter': None,
    }

    def instantiate(self, **hyperparams) -> 'FitCountDistributedInstance':
        return FitCountDistributedInstance(parent=self, **hyperparams)


class FitCountDistributedAlgorithmNeighbor(FakeDistributedAlgorithmNeighbor):
    '''A fake distributed algorithm neighbor state that counts the number of fits.'''
    fit_count: int

    def __init__(self, fit_count: int = 0, **kwargs):
        super().__init__(**kwargs)
        self.fit_count = fit_count

    def encode(self) -> bytes:
        '''Encode a message for distributed neighbors. '''
        return pickle.dumps(self.fit_count)

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'FitCountDistributedAlgorithmNeighbor':
        '''Decode a message from distributed neighbors.'''
        fit_count = pickle.loads(serialized_model)
        return cls(fit_count=fit_count)

    def state_differs(self,
                      distributed: DistributedConfig,
                      other: Optional[Any]) -> bool:
        '''Always report that our state has changed until we've been fit 3 times'''
        log.debug('Checking if state changed, fit count: %s', self.fit_count)
        _ = distributed
        _ = other
        return self.fit_count < 3


class FitCountDistributedInstance(FakeDistributedInstance):
    '''An instance of FakeAlgorithm for tests that counts fits.'''

    def _decode(self, serialized_model: bytes) -> FitCountDistributedAlgorithmNeighbor:
        '''Decode a message from distributed neighbors. '''
        return FitCountDistributedAlgorithmNeighbor.decode(serialized_model)

    def _fit(self, dataset: Optional[Dataset], **kwargs) -> None:
        '''Fit the model to the data. This is the actual implementation of fit.'''
        _ = dataset
        if self._my_state is None:
            self._my_state = FitCountDistributedAlgorithmNeighbor(fit_count=0)
        assert isinstance(self._my_state, FitCountDistributedAlgorithmNeighbor)
        self._my_state.fit_count += 1
        log.debug('Fit count incrementing to %s', self._my_state.fit_count)



'''Test the autoloading template catalog.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .parallel_step import ParallelStep
from .query_step import QueryStep
from .template_auto import TemplateCatalogAuto

# pylint: disable=missing-function-docstring,duplicate-code


def test_sunny_day() -> None:
    dut = TemplateCatalogAuto()
    demo_template = dut.lookup_by_name('tabular_classification')
    assert isinstance(demo_template.steps[1], ParallelStep)
    subpipe_keys = sorted(list(demo_template.steps[1].subpipeline_keys))
    assert ['attributes', 'target'] == subpipe_keys
    assert isinstance(demo_template.steps[3], QueryStep)


def test_lookup_by_datatype() -> None:
    dut = TemplateCatalogAuto()
    templates = dut.lookup_by_datatype(data_type='tabular')
    assert 'tabular_classification' in templates


def test_lookup_by_task() -> None:
    dut = TemplateCatalogAuto()
    templates = dut.lookup_by_task(task='binary_classification')
    assert 'tabular_classification' in templates


def test_lookup_by_both() -> None:
    dut = TemplateCatalogAuto()
    templates = dut.lookup_by_both(
        data_type='tabular',
        task='binary_classification')
    assert 'tabular_classification' in templates



'''Tests for wrangler_base.py'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code
# pylint: disable=missing-class-docstring,protected-access


from typing import Dict, Optional, Tuple
import numpy as np
import pandas as pd
import pytest

from ..executor.executor import ExecutorStub
from ..executor.simple.simple_executor import SimpleExecutor
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import (
    ExecutablePipeline, ExecutablePipelineStub, PipelineResult)
from ..metrics.impl.metric import MetricInvalidDatasetError
from ..problem_def.problem_def import ProblemDefinition
from ..tables.impl.table import TableFactory
from ..tables.impl.table_auto import TableCatalogAuto
from ..templates.impl.pipeline_template import PipelineTemplate
from ..templates.impl.template import TemplateCatalog

from .dataset import Column, Dataset, Metadata, RoleName
from .wrangler_base import WranglerBase, WranglerFailure

_ = TableCatalogAuto()  # pylint: disable=pointless-statement


# fakes -----------------------------------------------------

class FakeExecutor(ExecutorStub):

    def predict(self,
                dataset: Dataset,
                pipelines: Dict[Designator, ExecutablePipeline]
                ) -> Dict[Designator, PipelineResult]:
        '''Use a list of pipelines to predict from a dataset.'''
        predictions = dataset.output()
        # create a dataframe with the shape we expect predictions to be:
        # 1 column (the target) and a number of rows equal to the size of the input
        predictions.predictions = pd.DataFrame({
            dataset.metadata.roles[RoleName.TARGET][0].name:
                range(0, dataset.dataframe.shape[0])})
        return {d: PipelineResult(executable_pipeline=p, prediction=predictions)
                for d, p in pipelines.items()}


class FakeTemplateCatalog(TemplateCatalog):
    '''fake, no templates'''


class FakeWrangler(WranglerBase):

    def lookup_templates(self) -> Dict[str, PipelineTemplate]:
        return {}


# data -----------------------------------------------------


@pytest.fixture(scope='session')
def classification_results() -> Tuple[Dataset, Dataset]:
    metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        pos_labels={RoleName.TARGET: 'a'})
    pred = Dataset(metadata=metadata)
    pred.predictions = pd.DataFrame({'target': ['a', 'b', 'b', 'a']})
    pred.probabilities = TableFactory({'target': [.3, .4, .5, .5]})

    ground_truth = Dataset(metadata=metadata)
    ground_truth.ground_truth = pd.DataFrame({'target': ['a', 'a', 'b', 'a']})
    return (pred, ground_truth)


def get_data() -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
    data = pd.DataFrame(
        {
            'a': range(1, 21),
            'b': range(1, 21),
            'c': range(1, 21)
        }
    )
    return (data, data)

# problem def -----------------------------------------------------


def memory_problem_def() -> ProblemDefinition:
    return ProblemDefinition(clause={
        "dataset": {
            "config": "memory",
            "column_roles": {
                "target": {
                    "name": "c"
                }
            },
            "params": {
                "train_data": "train_data",
                "test_data": "test_data"
            }
        },
        "problem_type": {
            "task": "regression"
        }
    })

# tests ----------------------------------------------------------


def test_init_default() -> None:
    '''Test the components that wrangler defaults to.'''
    (train_data, test_data) = get_data()  # noqa: F841  # pylint: disable=unused-variable
    dut = FakeWrangler(problem_definition=memory_problem_def())
    assert dut._generator.__class__.__name__ == "GeneratorImpl"
    assert dut._template_catalog.__class__.__name__ == "TemplateCatalogAuto"


def test_init_custom() -> None:
    '''Test that overriding components when initializing the wrangler works.'''
    (train_data, test_data) = get_data()  # noqa: F841  # pylint: disable=unused-variable
    dut = FakeWrangler(
        problem_definition=memory_problem_def(),
        executor=FakeExecutor,
        template_catalog=FakeTemplateCatalog)
    assert dut._executor.__class__.__name__ == "FakeExecutor"
    assert dut._template_catalog.__class__.__name__ == "FakeTemplateCatalog"

    # Testing that non-set arguments are still default
    assert dut._generator.__class__.__name__ == "GeneratorImpl"


def test_wrangler_dataset() -> None:
    '''Test wrangler.dataset(), which takes a dataframe and returns a
    dataset containing the dataframe with metadata matching the problem definition.'''
    (train_data, test_data) = get_data()  # noqa: F841 pylint: disable=unused-variable
    problem_def = memory_problem_def()
    new_data = {
        'a': range(1, 31),
        'b': range(1, 31),
        'c': range(1, 31),
    }

    dut = FakeWrangler(
        problem_definition=problem_def,
        executor=FakeExecutor)

    got = dut.dataset(new_data)

    # Confirm that the dataframe matches
    pd.testing.assert_frame_equal(
        got.dataframe,
        pd.DataFrame(new_data))

    # Confirm that at least part of the metadata matches.
    assert set(got.metadata.roles[RoleName.ATTRIBUTE]) == {Column('a'), Column('b')}


def test_predict_bad_executor_kind() -> None:
    (train_data, test_data) = get_data()  # pylint: disable=unused-variable
    problem_def = memory_problem_def()
    executor = SimpleExecutor
    dut = FakeWrangler(
        problem_definition=problem_def,
        executor=executor)

    fake_des = Designator('fake_des')
    stub_pipeline = ExecutablePipelineStub(trained=True)

    with pytest.raises(WranglerFailure, match=r'(fake_des.*simple)|(simple.*fake_des)/i'):
        dut.predict(
            new_data=Dataset(),
            trained_pipelines={fake_des: stub_pipeline})


def test_predict_untrained_pipeline() -> None:
    (train_data, test_data) = get_data()  # pylint: disable=unused-variable
    problem_def = memory_problem_def()
    executor = FakeExecutor  # executor kind is 'stub_executor_kind', should match stub pipeline.
    dut = FakeWrangler(
        problem_definition=problem_def,
        executor=executor)

    fake_des = Designator('fake_des')
    stub_pipeline = ExecutablePipelineStub(trained=False)

    with pytest.raises(WranglerFailure, match=r'(fake_des.*trained)|(trained.*fake_des)/i'):
        dut.predict(
            new_data=Dataset(),
            trained_pipelines={fake_des: stub_pipeline})


def test_roc_curve(classification_results: Tuple[Dataset, Dataset]) -> None:
    pred, ground_truth = classification_results
    pred_no_proba = Dataset(pred, metadata=pred.metadata)
    del pred_no_proba['probabilities']
    dut = FakeWrangler(problem_definition=ProblemDefinition({
        "dataset": {
            "config": "ignore"
        },
        "problem_type": {
            "task": "test_task"
        },
    }))
    designator1 = Designator("fake_algorithm1")
    designator2 = Designator("fake_algorithm2")
    designator3 = Designator("fake_algorithm3")
    results: Dict[Designator, Optional[Dataset]] = {
        designator1: pred,
        designator2: None,
        designator3: pred_no_proba
    }
    got = dut.roc_curve(results=results, ground_truth=ground_truth)
    want = pd.DataFrame({
        'fpr': [0.0, 1.0 / 3.0, 1.0],
        'tpr': [0.0, 1.0, 1.0],
        'thresholds': [np.inf, 0.5, 0.3]})

    got_designator1 = got[designator1]
    assert isinstance(got_designator1, Dataset)
    pd.testing.assert_frame_equal(got_designator1.dataframe, want)

    got_designator2 = got[designator2]
    assert got_designator2 is None

    got_designator3 = got[designator3]
    assert isinstance(got_designator3, MetricInvalidDatasetError)



''''Wrap up the primary operations of rank aggregation.'''

from typing import Any, Dict, List

from ...instantiator.executable_pipeline import PipelineResults
from ...ranker.ranker import Ranking, Rankings

from .aggregator import AggregatorRanking, Aggregator


class AggregateRanker():
    '''Wrap up the primary operations of rank aggregation.'''
    _methods: List[Aggregator]
    _rankings: Rankings
    _results: PipelineResults

    def __init__(
            self,
            methods: List[Aggregator],
            rankings: Rankings,
            results: PipelineResults,
            **overrides: Any):
        self._methods = methods
        self._rankings = rankings
        self._results = results
        self._hyperparams = overrides

    def __call__(self) -> Dict[str, Ranking]:
        '''Testable component for rankings.'''
        new_rankings: Dict[str, Ranking] = {}
        for aggregation_method in self._methods:
            try:
                agg_ranking = aggregation_method.aggregate(self._rankings, **self._hyperparams)
            except Exception as err:  # pylint: disable=broad-exception-caught
                agg_ranking = AggregatorRanking(
                    metric=aggregation_method,
                    results=self._results,
                    scores=str(err))
            new_rankings[agg_ranking.metric.name] = agg_ranking
        return new_rankings



'''Communicator for tests where we don't communicate.'''

from queue import Queue
from typing import List, Optional, Set

from ..config_components.distributed_config import DistributedConfig
from ..neighbor_manager.node_id import NodeID
from ..wrangler.constants import Defaults

from .impl.communicator import Communicator
from .impl.communicator_catalog import CommunicatorCatalog


class CommunicatorStub(Communicator):
    '''Stub Communicator.'''
    name = 'stub_communicator'
    tags = {}

    def __init__(self,
                 distributed: Optional[DistributedConfig] = None,
                 known_neighbors: Optional[Set[NodeID]] = None,
                 my_id: Optional[NodeID] = None):
        node = my_id or NodeID(0)
        super().__init__(my_id=node,
                         known_neighbors=set(),
                         distributed=DistributedConfig(clause={}))

    def start(self,
              queue: Queue,
              timeout: float = Defaults.LISTENER_TIMEOUT
              ) -> None:
        pass

    def stop(self) -> None:
        pass

    def send(self,
             dest_ids: List[NodeID],
             payload: bytes) -> int:
        return 0

    def send_all(self, payload: bytes) -> int:
        return 0


def register(catalog: CommunicatorCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(CommunicatorStub)



'''Generic connection endpoint.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class Endpoint():
    '''Generic communication endpoint.'''



'''Decider that uses the algorithm-specific distance metric.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
import math
from typing import Mapping

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DecidersConfig, DeciderConfig
from ..neighbor_manager.node_id import NodeID  # type: ignore[import-untyped]

from .impl.decider import Decider
from .impl.decider_catalog import DeciderCatalog


class MaxDistanceDeciderConfig(DeciderConfig):
    '''Configuration for DistanceDecider.'''

    def __init__(self, name: str, clause: dict) -> None:
        super().__init__(name=name, clause=clause)

    def __str__(self) -> str:
        return f'num_neighbors: {self.num_neighbors}'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MaxDistanceDeciderConfig):
            return False
        return self.num_neighbors == other.num_neighbors

    class Defaults(Enum):
        '''Default values for the clause.'''
        NUM_NEIGHBORS = 1

    class Keys(AEnum):
        '''Keys for the clause.'''
        ENABLED = 'enabled'
        NUM_NEIGHBORS = 'num_neighbors'

    @property
    def num_neighbors(self) -> int:
        '''Get the number of most distant neighbors we would send to.'''
        return self._get_with_default(
            self.Keys.NUM_NEIGHBORS, dflt=self.Defaults.NUM_NEIGHBORS.value)


class MaxDistanceDecider(Decider):
    '''Decider that uses the algorithm-specific distance metric.'''
    name = 'max_distance'
    tags = {}
    _config: MaxDistanceDeciderConfig

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        assert my_meta.current_state is not None
        distances = []
        for nid in neighbors:
            if neighbors[nid].current_state is None:
                # A neighbor with no state is infinitely far away.
                distances.append((nid, math.inf))
                continue
            distance = my_meta.current_state.distance(neighbors[nid].current_state)
            distances.append((nid, distance))

        distances.sort(key=lambda x: x[1], reverse=True)

        if not distances:
            return False

        max_distance = distances[0][1]

        possible_neighbors = [nid for nid, distance in distances if distance == max_distance]

        if len(possible_neighbors) < self._config.num_neighbors:
            possible_neighbors = [
                nid for nid, _ in distances[:self._config.num_neighbors]]

        return neighbor_id in possible_neighbors


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(MaxDistanceDecider)
    DecidersConfig.register(name=MaxDistanceDecider.name, config_type=MaxDistanceDeciderConfig)



'''Tests for the StubDecider class.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..config_components.distributed_config import DecidersConfig, DistributedConfig
from .impl.decider_auto import DeciderCatalogAuto
from .stub_decider import DeciderStub


def test_sunny_day():
    '''Test the StubDecider class.'''
    _ = DeciderCatalogAuto()
    distributed = DistributedConfig(clause={
        'decider': {
            'stub_decider': {}
        }
    })
    decider = DeciderStub(config=distributed.decider.stub_decider)  # pylint: disable=no-member
    assert decider.decide(None, None, None) is True


def test_catalog():
    '''Test the decider_stub in the DeciderCatalog class.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'stub_decider': {}
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider = catalog.lookup_by_name('stub_decider')(config=distributed.decider.stub_decider)  # pylint: disable=no-member
    assert decider.decide(None, None, None) is True

    assert 'stub_decider' in [k.value for k in DecidersConfig.Keys]



'''A Stub Table.'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Dict, List, Union

from .impl.table import TableFactory, Table
from .impl.table_catalog import TableCatalog


class StubType():
    '''A stub type.'''
    def __init__(self, token: str = 'stub'):
        self._token = token

    @property
    def token(self):
        '''Return the token.'''
        return self._token


class StubTable(Table):
    '''A Table representing a pandas Stub.'''
    name = 'stub'
    _value: StubType

    def to_csv(self, *args, **kwargs) -> str:
        '''Output the table as a CSV string.'''
        return 'stub'

    def as_dict(self, *args, **kwargs) -> Dict[Union[str, int], Any]:
        '''Convert table to a dictionary.'''
        return {'stub': StubType()}

    def drop(self, columns: List[Union[int, str]]) -> Table:
        '''Return a copy of this table with selected columns dropped.'''
        return self

    @property
    def columns(self) -> List[Union[int, str]]:
        '''Return the column names of the table.'''
        return []

    @columns.setter
    def columns(self, value: List[Union[int, str]]):
        '''Set the column names of the table.'''
        _ = value

    @property
    def empty(self):
        '''Return True if the table is empty.'''
        return True


def register(catalog: TableCatalog):
    '''Register the StubTable with the catalog.'''
    catalog.register(StubTable)
    TableFactory.register_constructor(StubType, StubTable)
    TableFactory.register_exporter(from_type=StubType, to_type=StubType, exporter=StubTable.value)
